"""UDC16 pedestals generatior.

No blocks, readout the entire 64 widnows in one sweep.
"""

import logging
import time
from collections import deque

import numpy as np

from naludaq.controllers import get_board_controller
from naludaq.helpers.exceptions import PedestalsDataCaptureError
from naludaq.tools.pedestals.pedestals_controller import (
    PedestalsController,
    _update_progress,
    increase_margin,
    sleep_calc,
)

LOGGER = logging.getLogger("naludaq.udc16_pedestal_generator")


class UDC16PedestalsGenerator(PedestalsController):
    def __init__(
        self,
        board,
        num_captures: int = 10,
        num_warmup_events: int = 10,
        channels: "list[int]" = None,
    ):
        super().__init__(
            board, num_captures, channels=None
        )  # Cannot select channels on UDC16

        self.sleeptime = 2.5  # sleep_calc(self.board, self.margin)
        self.num_warmup_events = 0

    def _capture_data_for_pedestals(
        self,
        num_captures: int = None,
        num_warmup_events: int = 0,
        channels: list = None,
        retries: int = 10,
    ) -> list:
        """Read the data for pedestals generation.

        Reads several samples from each block to average the data.
        This data is stored in the pedstore after beging processed in the
        data processing pipeline.

        Tells the board to capture events.

        Args:
            num_captures (int): number of events used to generate pedestals.
            warmup_events(int): number of warmup captures before events used for pedestals.
            channels (list): list of channels to capture for, or `None`
                to readout all channels
            retries (int): amount of retries if no valid data is captured
        """
        if num_captures is None:
            num_captures = self.num_captures
        self.num_warmup_events = num_warmup_events
        total_amount = num_captures + num_warmup_events
        needed_amount = total_amount
        prev_amount = -1
        tcnt = 0
        buffer = deque()

        LOGGER.debug("Capturing data for pedestals")
        # STATUS UPDATES ####################################
        min_value = 0
        max_value = 90
        step_value = (max_value - min_value) / total_amount
        #####################################################

        self._daq.start_capture()

        while len(buffer) < total_amount and not self._cancel:
            needed_amount = max(total_amount - len(buffer), 1) + 1
            LOGGER.debug("New run: %s", needed_amount)
            initial_data = self._get_data(
                needed_amount,
                prog_min=step_value * len(buffer),
                prog_max=max_value,
            )
            # When running in a thread, cancel to stop
            if self._cancel:
                return

            # validate captured data
            validated_data = deque()
            for evt in initial_data:
                if self._validate_event(evt):
                    validated_data.append(evt)
            LOGGER.debug(f"Validated: {len(validated_data)}/{len(initial_data)}")

            buffer.extend(validated_data)

            if len(initial_data) < needed_amount:
                self._clear_input_buffer()

            # No valid data captured
            if len(buffer) == prev_amount:
                tcnt += 1
                if tcnt >= retries:
                    break
            prev_amount = len(buffer)

        if len(buffer) < total_amount:  # This happends on timeout
            raise PedestalsDataCaptureError(
                f"Not enough data captured {len(buffer)}/{total_amount}"
            )

        pedestals_data = list(buffer)[
            -self.num_captures :
        ]  # deque doesn't support indexing

        # self.validated_data is used to generate pedestals
        self.validated_data = pedestals_data

    def _clear_input_buffer(self):
        """Make sure all data is cleared from all buffers"""
        while self.board.connection.in_waiting:
            self.board.connection.reset_input_buffer()
            time.sleep(0.1)

    def _get_data(self, amount: int, prog_min: int = 0, prog_max: int = 90):
        """Try to capture the amount of even, no amount guaranteed.

        This function fires `amount` software triggers to the board and waits
        enough time to receive them. Doesn't guarantee the data though and
        validation should be used.

        Keep track of progress using the prog_min and _max

        Args:
            amount (int): desired amount of data
            prog_min (int): minumum progress
            prog_max (int): maximum progress

        Returns:
            collections.deque with the events captured
        """
        prog_step = (prog_max - prog_min) / (amount or 1)

        self.brd_ctrl = get_board_controller(self.board)
        self._clear_input_buffer()

        self._start_capture()
        time.sleep(0.1)
        # Replace with bestest code! #############################################################
        for idx in range(amount):
            _update_progress(
                self._progress,
                int(prog_min + prog_step * (idx + 1)),
                "Capturing data...",
            )
            time.sleep(0.1)
            self.brd_ctrl.toggle_trigger()
            time.sleep(self.sleeptime)
            if self._cancel:
                break

        self._stop_capture()

        return self._daq.output_buffer

    def _start_capture(self):
        """Start capturing data"""
        # self.board.connection.reset_input_buffer()
        # self.brd_ctrl.sysrst()
        while self.board.connection.in_waiting:
            self.board.connection.reset_input_buffer()
            time.sleep(0.1)
        self._daq.output_buffer = deque()
        self._daq.start_capture()
        self.brd_ctrl.toggle_trigger()

    def _stop_capture(self):
        """Stop capturing data"""
        self._daq.stop_capture()
        self._daq.stop_workers()

    def _backup_settings(self) -> dict:
        """Store relevant registers"""
        return {}

    def _restore_backup_settings(self, backup):
        """Restore backed up registers"""
        pass

    def _set_pedestals_raw_to_zero(self, num_captures):
        brd = self.board
        output = np.zeros(
            shape=(
                brd.params.get("channels", 16),
                brd.params.get("windows", 64),
                brd.params.get("samples", 64),
                num_captures,
            )
        )

        output[:] = np.NaN
        return output

    def _validate_event(self, event):
        """Check if the event has a data field, which means it's parsed"""
        try:
            _ = event["data"]
        except KeyError:
            return False
        return True

    def _generate_pedestals_data(self, rawdata=None, num_captures=None):
        """Generates the pedestals data from captured data.

        Uses captured events in the pedestals_raw_data deque
        to generate the data.

        Args:
            nCaptures (int): Number of captures, default 10
            blocksize (int): size of blocks, default 16
        """
        _update_progress(self._progress, 95, "Generating pedestals...")

        if num_captures is None:
            num_captures = self.num_captures
        if rawdata is None:
            rawdata = self.validated_data

        block_size = 1

        LOGGER.debug(
            "Generating pedestals from %s samples with blocksize: %s",
            num_captures,
            block_size,
        )

        return self._validate_and_transfer_data_to_pedestals(
            rawdata, block_size, num_captures
        )

    def _validate_and_transfer_data_to_pedestals(
        self, validated_data, block_size, num_captures
    ):
        """Move data to board.pedestals after validating.

        Since data already is prevalidated this is just a logic check,

        Args:
            block_size(int): windows per block
            num_captures(int): amount of events averaged.
        """
        board = self.board
        channels = board.params["channels"]
        samples = board.params["samples"]

        for cap, event in enumerate(validated_data[-num_captures:]):
            for chan in range(channels):
                for window in event["window_labels"][chan]:
                    for sample in range(samples):
                        real_window = list(event["window_labels"][chan]).index(window)
                        index = int(sample + real_window * samples)
                        data = event["data"][chan][index]
                        board.pedestals["rawdata"][chan][window][sample][cap] = data

        return True

    def _reset_pedestals_data(self, num_captures=None):
        """Reset all pedestals data to an empty shape.

        The shape of the empty pedestals.data is:
        [channels][windows][samples][num_captures]

        Args:
            num_captures (int): Override number of captures for pedestals generation.
                Leave as None is unsure.
        """
        if num_captures is None:
            num_captures = self.num_captures
        pedestals = self.board.pedestals
        if pedestals is None:
            pedestals = dict()
        pedestals["rawdata"] = self._set_pedestals_raw_to_zero(num_captures)
        self.board.pedestals = pedestals

    def _generate_pedestals_from_data(self):
        """Generate pedestals data and average it.

        Takes the generated pedestals data and generates the actual pedestals.
        The pedestals are stored in the controller and can be retreived with
        get_pedestals().

        """
        pedestals = self.board.pedestals
        rawdata = pedestals["rawdata"]
        output = np.nanmean(rawdata, axis=3)
        np.nan_to_num(output, nan=0, copy=False)
        pedestals["data"] = output

    def _store_time_metadata(self):
        """Store event times into pedestals metadata"""
        LOGGER.debug("Storing time metadata")
        self.metadata.store_event_times(self.validated_data)

    def _increase_sleeptime(self, needed_amount, captured_amount):
        """Increase the sleeptime based on received vs expected.

        Formula is intended to change the margin incrementally without
        causing to much oscillation.
        """
        # inc = (needed_amount / len(self.block_buffer))/2 + 1 settles at 0.45
        # inc = (needed_amount / len(self.block_buffer))/2 fails
        # inc = (needed_amount / len(self.block_buffer)) Oscillates
        block_len = max(captured_amount, 0.1)
        inc = (needed_amount / block_len) / 3 + 1  # Settles at 0.38
        self.margin = increase_margin(self.margin, margin_inc=inc, max_margin=10)
        self.sleeptime = sleep_calc(self.board, self.margin)
        LOGGER.debug("Waittime changed to %s", self.sleeptime)



"""Driver for the LTC2990 I2C voltage/current/temperature monitor.

Datasheet:
    https://www.analog.com/media/en/technical-documentation/data-sheets/ltc2990.pdf

Mitchell Matsumori-Kelly
"""
import logging
import time

from naludaq.helpers.exceptions import BadDataError, I2CError
from naludaq.helpers.helper_functions import type_name

from .i2c_device import I2CDevice

logger = logging.getLogger("naludaq.ltc2990")


class LTC2990(I2CDevice):
    # bitmasks
    CONTROL_MEASURE_ALL = 0b000_11_000
    CONTROL_VOLTAGE_ALL = 0b000_00_111

    # register map
    REGISTER_STATUS = 0x00
    REGISTER_CONTROL = 0x01
    REGISTER_TRIGGER = 0x02
    REGISTER_V1_MSB = 0x06
    REGISTER_V2_MSB = 0x08
    REGISTER_V3_MSB = 0x0A
    REGISTER_V4_MSB = 0x0C
    REGISTER_VCC_MSB = 0x0E
    VOLTAGE_REGISTER_MAP = {
        1: REGISTER_V1_MSB,
        2: REGISTER_V2_MSB,
        3: REGISTER_V3_MSB,
        4: REGISTER_V4_MSB,
    }

    # values relevant to calculations
    SE_VOLTAGE_FACTOR = 305.18e-6

    def __init__(self, board, addr: int):
        """Driver for the LTC2990 I2C voltage/current/temperature
        monitor.

        On instantiation will reset the operating mode to the default
        MEASURE_ALL and CONTROL_ALL.

        Attributes:
            board (Board): The board object. Must have an open connection.
            addr (int): Address of the device.

        Raises:
            I2CError: if the device could not be initialized.
        """
        super().__init__(board, addr)
        self._log_prefix = f"[LTC2990 at {addr:02X}]"

        try:
            self.set_mode()  # Reset to default mode on initialization.
        except I2CError as e:
            raise I2CError("Could not initialize the device") from e

    def set_mode(self, mode=CONTROL_MEASURE_ALL | CONTROL_VOLTAGE_ALL):
        """Sets the mode of operation for the device.

        Args:
            mode (int): the mode, bits [4:0] of the control register.
                See page 16 of the datasheet.

        Raises:
            I2CError: if the deivce did not acknowledge the command
        """
        logging.debug(f"{self._log_prefix} - setting mode to {mode:08b}")
        self.write_register(LTC2990.REGISTER_CONTROL, bytes([mode]))

    def trigger_conversion(self):
        """Triggers a conversion operation in the device.

        It is necessary to call this method:
            - Before _every_ read of a measurement if in single-shot mode
            - Once initially if in continuous mode.

        Raises:
            I2CError: if the deivce did not acknowledge the command
        """
        logging.debug(f"{self._log_prefix} - triggering conversion")
        self.write_register(LTC2990.REGISTER_TRIGGER, bytes([1]))
        time.sleep(0.01)

    def read_voltage(self, channel: int, attempts=5) -> float:
        """Reads the voltage for a particular channel on the device.

        This function WILL fail if the conversion has not been properly
        triggered.

        Args:
            channel (int): the channel number, must be 1-4.
            attempts (int): number of times to try reading the data.
                The data comes back flagged as invalid if the device
                has not been triggered or is not finished converting.

        Returns:
            The voltage in volts.

        Raises:
            BadDataError if the device has not stored a converted value.
        """
        register = LTC2990.VOLTAGE_REGISTER_MAP.get(channel, None)
        if register is None:
            raise ValueError(f"Channel must be 1-4, not {channel}")
        try:
            voltage_reg = self._read_voltage_register(register, attempts)
        except BadDataError:
            raise
        voltage = self._convert_register_to_voltage(voltage_reg)
        logging.debug(
            f"{self._log_prefix} - channel {channel} read back as {voltage} V"
        )
        return voltage

    def read_vcc(self, attempts=5) -> float:
        """Reads the VCC supply voltage of the device.

        This function WILL fail if the conversion has not been properly
        triggered.

        Args:
            attempts (int): number of times to try reading the data.
                The data comes back flagged as invalid if the device
                has not been triggered or is not finished converting.

        Returns:
            The voltage in volts.

        Raises:
            BadDataError if the device has not stored a converted value.
        """
        try:
            register = LTC2990.REGISTER_VCC_MSB
            voltage_reg = self._read_voltage_register(register, attempts)
        except BadDataError:
            raise
        self._fix_device_state()  # weird stuff happens when the register pointer rolls over?

        voltage = self._convert_vcc_register_to_voltage(voltage_reg)
        logging.debug(f"{self._log_prefix} - VCC read back as {voltage} V")
        return voltage

    def _read_voltage_register(self, register, attempts=5) -> int:
        """Reads the voltage register for the given channel.
        Make sure a conversion has been triggered before reading!

        Attempts to read the voltage several times until the data
        is flagged as valid. Otherwise it raises an error if the
        number of attempts has been exceeded.

        Args:
            channel (int): the channel to read from
            attempts (int): the number of times to read until
                valid data is received.

        Returns:
            int: The value of the register.

        Raises:
            BadDataError: if the read operation failed.
        """
        for _ in range(attempts):
            try:
                # Reading one 1-byte wide register results in two bytes thrown back...
                msb, lsb = self.read_register(register, width=2)[0:2]
            except:
                continue
            data_valid = bool(msb >> 7)
            if data_valid:
                break
            logging.debug(
                f"{self._log_prefix} - failed to fetch valid data, retrying..."
            )
            time.sleep(0.1)
        else:
            raise BadDataError("No valid data returned from LTC2990")

        return (msb << 8) | lsb

    def _fix_device_state(self):
        """For whatever reason the device gets into an invalid state after reading
        the VCC registers. Right after the VCC registers are read, no write operation
        will be acknowledged by the device.

        This function performs a dummy writes to the status register (ignored because
        it's read-only) which seems to fix the issue.
        """
        time.sleep(0.05)
        for _ in range(2):
            try:
                self.write_register(LTC2990.REGISTER_STATUS, bytes([0]))
            except:
                pass

    @staticmethod
    def _convert_vcc_register_to_voltage(register_value: int):
        """Converts the value held in the VCC voltage register to
        the actual voltage.

        Args:
            register_value (int): the value of the register.

        Returns:
            The value in Volts.

        Raises:
            TypeError if the value is not an `int`.
            ValueError if the value is out of bounds for a two-byte unsigned int.
        """
        try:
            return 2.5 + LTC2990._convert_register_to_voltage(register_value)
        except (ValueError, TypeError):
            raise

    @staticmethod
    def _convert_register_to_voltage(register_value: int):
        """Converts the value held in a pair of voltage registers to
        the actual voltage.

        Args:
            register_value (int): the value of the register
                including the MSB and LSB.

        Returns:
            The value in Volts.

        Raises:
            TypeError if the value is not an `int`.
            ValueError if the value is out of bounds for a two-byte unsigned int.
        """
        if not isinstance(register_value, int):
            raise TypeError(
                f"Register value must be an int, not {type_name(register_value)}"
            )
        if not 0 <= register_value <= 0xFFFF:  # voltage registers are 2 bytes
            raise ValueError(
                f"Register value {register_value:X} is out of bounds (0-0xFFFF)"
            )

        data_mask = 0x3FFF
        data_bits = register_value & data_mask
        sign_bit = (register_value >> 14) & 1

        factor = LTC2990.SE_VOLTAGE_FACTOR
        if sign_bit == 1:
            data_bits = -(
                (data_mask ^ data_bits) + 1
            )  # register is stored in two's complement
        voltage = data_bits * factor
        return voltage



from .aardvarcv3 import Aardvarcv3ReadoutController
from .default import ReadoutController
from .hdsoc import HDSoCReadoutController
from .trbhm import TrbhmReadoutController


def get_readout_controller(board):
    """Gets the readout controller which is appropriate for the given board.

    hdsocv1 -> HDSoCReadoutController
    aardvarcv3 => Aardvarcv3ReadoutController
    default -> ReadoutController

    Args:
        board (Board): the board object

    Returns:
        The readout controller.
    """
    return {
        "hdsocv1": HDSoCReadoutController,
        "hdsocv1_evalr1": HDSoCReadoutController,
        "hdsocv1_evalr2": HDSoCReadoutController,
        "aardvarcv3": Aardvarcv3ReadoutController,
        "aardvarcv4": Aardvarcv3ReadoutController,
        "trbhm": TrbhmReadoutController,
        "aodsoc_aods": TrbhmReadoutController,
        "aodsoc_asoc": TrbhmReadoutController,
    }.get(board.model, ReadoutController)(board)



import copy
from dataclasses import dataclass

from naluconfigs import get_available_models

from naludaq.helpers import type_name


@dataclass()
class Feature:
    """Dataclass describing a feature"""

    description: str
    long_description: str
    beta_models: list[str]


_FEATURES = {
    "adc2mv": Feature(
        description="ADC to mV Calibration",
        long_description="Allows calibration to convert from units of ADC to mV",
        beta_models=get_available_models(),
    ),
    "dac_sweep": Feature(
        description="External DAC Sweep",
        long_description=(
            "Gathers data at various DAC values to characterize the response to "
            "different input levels"
        ),
        beta_models=get_available_models(),
    ),
    "ext_dac": Feature(
        description="External DAC",
        long_description="Allows control of the external DAC on the board",
        beta_models=[],
    ),
    "pedestals": Feature(
        description="Pedestal Calibration",
        long_description="Allows generation of pedestals data and subtraction of built-in noise",
        beta_models=[],
    ),
    "threshold_scan": Feature(
        description="Threshold Scan",
        long_description=(
            "Scans over different trigger thresholds to determine the "
            "threshold at which the ASIC will trigger on a signal"
        ),
        beta_models=[],
    ),
    "tia_dac": Feature(
        description="TIA DAC",
        long_description=(
            "Enables use of the transimpedence amplifier (not present on all models) "
            "to digitize an input current rather than input voltage"
        ),
        beta_models=[],
    ),
    "timing_calibration": Feature(
        description="Timing Calibration",
        long_description=(
            "Enables calibration of the internal timing circuit "
            "for better sample-to-sample timing precision"
        ),
        beta_models=get_available_models(),
    ),
    "gain_stage_tuner": Feature(
        description="Gain Stage Tuner",
        long_description=(
            "Calibrates external dac and ISEL for a specific "
            "gain stage configuration"
        ),
        beta_models=get_available_models(),
    ),
}


def features_names() -> list[str]:
    """Get a list of available feature names"""
    return list(_FEATURES.keys())


def all_feature_info() -> dict[str, Feature]:
    """Get a dictionary of all feature information"""
    return copy.deepcopy(_FEATURES)


def feature_info(feature_name: str) -> Feature:
    """Get information about about a feature"""
    _validate_feature_name_or_raise(feature_name)
    return copy.deepcopy(_FEATURES[feature_name])


def is_beta(feature_name: str, model: str) -> bool:
    """Checks whether a feature is currently beta for a given board model"""
    _validate_model_or_raise(model)
    return model in feature_info(feature_name).beta_models


def is_available(feature_name: str, model: str, board_features: dict) -> bool:
    """Checks whether the feature is available.

    Available means that the feature switch is `True` in the configs,
    and if it is a beta feature, that beta features are enabled.

    Args:
        feature_name (dict): name of the feature.
        model (str): board model
        board_features (dict): feature dict from `board.features`.

    Returns:
        bool: True if the feature is available.
    """
    _validate_feature_dict_or_raise(board_features)
    available = board_features.get(feature_name, False)
    beta_enabled = board_features.get("beta_features", False)
    if is_beta(feature_name, model) and not beta_enabled:
        available = False
    return available


def _validate_feature_name_or_raise(feature_name: str):
    """Raise an error if the feature name is wrong type/value"""
    if not isinstance(feature_name, str):
        raise TypeError(f"Name must be a string, not {type_name(feature_name)}")
    if feature_name not in _FEATURES:
        raise ValueError(f"Invalid feature name '{feature_name}'")


def _validate_feature_dict_or_raise(board_features: dict):
    """Raise an error if the board features dict is the wrong type"""
    if not isinstance(board_features, dict):
        raise TypeError(
            f"Board features must be a dict, not {type_name(board_features)}"
        )


def _validate_model_or_raise(model: str):
    """Raise an error if the board model is the wrong type/value"""
    if not isinstance(model, str):
        raise TypeError(f"Board model must be a str, not {type_name(model)}")
    if model not in get_available_models():
        raise ValueError(f"Invalid board model '{model}'")



"""Pedestals generator for UPAC32.

UPAC32 works differently because the data cannot be captured in blocks.
"""
import copy
import logging
from collections import deque

from naludaq.controllers import get_board_controller
from naludaq.tools import EventWaiter
from naludaq.tools.pedestals.pedestals_controller import (
    PedestalsController,
    _update_progress,
    sleep_calc,
)

LOGGER = logging.getLogger("naludaq.upac32_pedestal_generator")


class UpacPedestalsController(PedestalsController):
    """Pedestals controller for UPAC32"""

    def __init__(
        self,
        board,
        num_captures: int = 10,
        num_warmup_events: int = 10,
        channels: "list[int]" = None,
    ):
        super().__init__(board, num_captures, num_warmup_events)
        self.failed_validation = deque()

        self.margin = 2
        self.sleeptime = sleep_calc(self.board, self.margin)

    def _capture_data_for_pedestals(self, timeout: int = 3):
        """ """
        num_captures = self.num_captures
        num_warmup_events = self.num_warmup_events
        expected_amount = num_captures + num_warmup_events

        timeout_cnt = 0
        self.block_buffer = deque(maxlen=expected_amount)
        self._daq.output_buffer = self.block_buffer

        while len(self.validated_data) < num_captures and self._cancel is False:
            if timeout_cnt > timeout:
                break
            self.block_buffer = deque(maxlen=expected_amount)
            self._daq.output_buffer = self.block_buffer

            # Check the amount of data needed
            needed_amount = expected_amount - len(self.validated_data)
            # Extra code for status #############################################
            min_value = 0
            max_value = 80
            step_value = (max_value - min_value) / expected_amount
            #####################################################################
            # Capture data
            self._daq.start_capture()
            get_board_controller(self.board).start_readout()
            for idx in range(needed_amount + 1):
                if self._cancel:
                    break
                required_buffer_size = len(self._daq.output_buffer) + 1
                get_board_controller(self.board).toggle_trigger()

                _update_progress(
                    self.progress,
                    min_value + step_value * idx,
                    f"Capturing data, block: {len(self.block_buffer)+1}/{expected_amount}",
                )

                waiter = EventWaiter(
                    self._daq.output_buffer,
                    amount=required_buffer_size,
                    timeout=self.sleeptime,
                )
                try:
                    waiter.start(blocking=True)
                except TimeoutError:
                    LOGGER.warning("Event timed out")
            get_board_controller(self.board).stop_readout()
            self._daq.stop_capture()

            if len(self.block_buffer) != needed_amount:
                LOGGER.warning("data capture failed, increasing sleeptime")
                self._increase_sleeptime(needed_amount)

            # transfer data
            while self.block_buffer and len(self.validated_data) != num_captures:
                try:
                    event = (
                        self.block_buffer.pop()
                    )  # Pop from right skips warmup events
                except Exception as emsg:
                    continue
                LOGGER.debug("Transfer event")
                _update_progress(
                    self.progress,
                    min_value + step_value * len(self.validated_data),
                    f"Validate data {len(self.validated_data) + 1}/{expected_amount}",
                )

                # transfer validated data
                if event.get("data", None) is not None:
                    self.validated_data.append(event)

            if len(self.validated_data) == 0:
                LOGGER.debug(
                    "Pedestals buffer is empty, indicating transmission error, retrying."
                )
                timeout_cnt += 1
                continue
            elif self._store_warmup_events:
                self._warmup_data.append(self.block_buffer)

            # if len(self.validated_data) != needed_amount:
            #     LOGGER.debug("data capture failed, restarting")
            #     self._increase_sleeptime(needed_amount)

    def _backup_settings(self) -> dict:
        """Backup settings that might get overwritten to a dict.

        Returns:
            dict with the backup settings
        """
        backup = {
            "control_registers": copy.deepcopy(
                self.board.registers.get("control_registers", {})
            ),
        }

        return backup

    def _restore_backup_settings(self, backup_settings):
        """Restore all backuped settings to the board.

        Returns:
            True if settings have been restored, False if no old settings were found.
        """
        if not backup_settings:
            return

        for register in ["control_registers"]:
            reg = backup_settings.get(register, {})
            if reg:
                self.board.registers[register] = reg

    def _generate_pedestals_data(self, num_captures=None):
        """Generates the pedestals data from captured data.

        Uses captured events in the pedestals_raw_data deque
        to generate the data.

        Args:
            nCaptures (int): Number of captures, default 10
            blocksize (int): size of blocks, default 16
        """

        if num_captures is None:
            num_captures = self.num_captures

        block_size = self.board.params["pedestals_blocks"]

        self._reset_pedestals_data(num_captures)

        LOGGER.debug(
            "Generating pedestals from %s samples with blocksize: %s",
            num_captures,
            block_size,
        )

        return self._validate_and_transfer_data_to_pedestals(block_size, num_captures)

    def _validate_and_transfer_data_to_pedestals(self, block_size, num_captures):
        """Move data to board.pedestals after validating.

        Since data already is prevalidated this is just a logic check,

        Args:
            block_size(int): windows per block
            num_captures(int): amount of events averaged.
        """
        num_blocks = self._calculate_number_blocks(
            block_size, self.board.params["windows"]
        )
        for block in range(num_blocks):
            LOGGER.info("Block %s/%s", block, num_blocks)

            for cap in range(num_captures):
                try:
                    event = self.validated_data.popleft()
                except IndexError:
                    return False

                for chan in range(self.board.params["channels"]):
                    for window_num in range(block_size):
                        real_window = event["window_labels"][chan].index(window_num)
                        window = window_num + block * block_size
                        if block != 0 and (window < block_size):
                            continue
                        if window >= self.board.params["windows"]:
                            continue
                        for sample in range(self.board.params["samples"]):
                            index = sample + real_window * self.board.params["samples"]
                            data = event["data"][chan][index]
                            self.board.pedestals["rawdata"][chan][window][sample][
                                cap
                            ] = data

        return True

    def _store_time_metadata(self):
        """Store event times into pedestals metadata"""
        LOGGER.debug("Storing time metadata")
        self.metadata.store_event_times(self.validated_data)



"""Digital Registers
====================
Registers controlling the digital part of the ASIC.

This is reliant on the digital_registers portion of the .yml file.

These registers control the digital side of the ASICs.
By changing these registers it's possible to
change how and what the ASIC is reading.

Command structure:
------------------
Updating a digital command uses a format 'B01AAVVV'

- B tells the FPGA command parser to expect a analog/digital command.
- M is mode, 0 = write, 1 = read
- AA = address
- VVV = value

Will be interpereted by the FPGA as hex. it's 32-bit = 8 hex values.

Example:
---------

.. code-block: python

    dc = DigitalRegisters(board)
    dc.read("regname")
    dc.write("regname", intvalue)

"""
from logging import getLogger

from naludaq.communication import registers

LOGGER = getLogger(__name__)


class DigitalRegisters(registers.ReadRegisters, registers.WriteRegisters):
    """Digital Registers

    Attributes:
        board:
    """

    def __init__(self, board):
        super().__init__(board, "digital_registers")
        # self.parse_answer = get_answer_parser()
        self._write_cmd = 0xB01
        self._read_cmd = 0xB42
        self.width_cmd = 3
        self.width_addr = 2  # hex char
        self.width_val = 3  # hex char

    def write_all(self):
        """Write all software registers to the hardware.

        Writes all registers on the board-object to the physical board.
        It will send all the commands one by one to the board.
        Effectively syncing the software registers with the board by
        sending all values in the software to the board.
        """
        addresses = self._multichip_fix(self.list_addresses())
        for addr in addresses:
            self._write_register_or_raise(addr)

    def _multichip_fix(self, addresses):
        """Fix for multichip boards.

        If the board is a multichip board, the chip_id register
        should not be written to the board.
        """
        if self.board.params.get("num_chips", 1) != 1:
            try:
                chip_id_addr = self._get_addr_from_name_or_raise("chip_id")
                addresses.remove(chip_id_addr)
            except ValueError:
                pass
        return addresses

    def _read_all_new(self):
        """Read all registers [BETA].

        Returns:
            dict: register values structured as {reg_name: reg dict}
        """
        # Digital registers have no command ID, so we need to read them synchronously
        values = {}
        for reg_addr in self.list_addresses():
            response = self._read_from_hardware(reg_addr)
            reg = self._parse_response_to_names(response, reg_addr)
            values.update(reg)
        return values

    @registers.clear_buffer
    def read_many(self, names: list[str]) -> list[str]:
        """Read several registers at once.

        Args:
            names (Iterable[str]): names of registers to read

        Returns:
            list[int]: list of register values in the same order as provided.
        """
        self._validate_names_or_raise(names)

        # Digital registers don't have command IDs so concurrent reads are dangerous
        return [self.read(name)["value"] for name in names]



import copy

from naludaq.communication import AnalogRegisters, ControlRegisters, DigitalRegisters
from naludaq.controllers import get_readout_controller


class BoardBackup:
    """A back up of the board registers at a point in time.
    Can be used to restore the board to an earlier state.
    """

    def __init__(self, board) -> None:
        self._board = board
        self._backup = None

    def backup(self):
        """Back up the registers for the board."""
        rc = get_readout_controller(self._board)
        self._backup = {
            "readout_channels": rc.get_readout_channels(),
            "registers": copy.deepcopy(self._board.registers),
        }

    def restore(self, exclude_registers: list[str] = []):
        """Restores the backup by writing all registers to the board.

        Args:
            exclude_registers (list[str]): list of register names to exclude
                from restoration (e.g. "analog_registers").

        Raises:
            ValueError: if there was no backup previously created.
        """
        if self._backup is None:
            raise ValueError("No backup was previously generated")
        rc = get_readout_controller(self._board)
        rc.set_readout_channels(self._backup["readout_channels"])

        reg_types = {
            "analog_registers": AnalogRegisters,
            "control_registers": ControlRegisters,
            "digital_registers": DigitalRegisters,
        }
        for name, type_ in reg_types.items():
            if name in exclude_registers:
                continue
            self._board.registers[name] = self._backup["registers"][name]
            type_(self._board).write_all()



# Builtin
from logging import getLogger

# 3rd Party
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score

# Internal
from naludaq.board import Board
from naludaq.helpers.helper_functions import type_name

LOGGER = getLogger(__name__)


class ADCLinearReg:
    """
    This class utilizes DAC sweep data collected from the dac sweep controller,
    and is used to calculate two values: linear region and linear regression.
    For linear region, it finds the largest region that is linear, defined by
    R-squared parameters. For linear regression, the slope and intercept  is
    calculated for each sample in each channel. Unless a linear region is
    defined, it will calculate the slope and intercept using the region closest
    to 600-1000 mV.

    Args:
        board (naludaq.board): Board object used for params.
        min_dac_val (int or list): The beginning of the linear region in mV
        max_dac_val (int or list): The ending of the linear region in mV

    Raises:
        TypeError if the board parameter is not a valid type
        InvalidBoardModelError if the provided str for the board model is invalid.

    """

    def __init__(self, board: Board):
        if not isinstance(board, Board):
            raise TypeError(f"Invalid board type. Got {type_name(board)}, expected Board")
        self.board = board
        self.channels = self.board.params["channels"]
        self.min_dac_val = None
        self.max_dac_val = None
        self._progress: list = []
        self._cancel = False

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError("Progress is stored in a list")
        self._progress = value

    def cancel(self):
        """Cancels the linear regression as soon as possible.
        No data is generated, and the `linear_regression` function
        will return `None`.

        Can only be called from a separate thread.
        """
        self._cancel = True

    def linear_regression(
        self, dac_data: dict, minimum_r2: float = 0.9998, minimum_range: int = 10
    ) -> dict:
        """Calculate the slope and intercept per sample using dac sweep data.

        Args:
            dac_data (dict): DAC sweep data collected from dac_sweep_controller
            minimum_r2 (float): The minimum R-squared value a region needs to be considered
                linear.
            minimum_range (int): The minimum number of points within the two endpoints to
                be considered a region. Used to optimize calculations of endpoints.
        Returns:
            lr_dict (dict): Contains slope & intercept of every sample per channel,
                formatted as:
                    {
                        "slope": 2D Array, channel major, sample minor,
                        "intercept": 2D Array, channel major, sample minor,
                        "linear_region": list of minmax tuples per channel
                    }
                Will return None if the operation was canceled.

        Raises:
            NoLinearRegionFoundError: No linear region was found for a channel.
        """
        self._cancel = False

        average_per_sample = list()
        sample_intercept = []
        sample_slope = []
        LOGGER.info("Reformatting DAC Sweep data")

        voltages = list(dac_data.keys())
        LOGGER.info("Voltages: %s", voltages)

        voltages = self._dac2mv(np.array(voltages))
        LOGGER.info("Converted voltages: %s", voltages)
        # Format data in form average_per_sample[channel #][sample #][dac_val]
        LOGGER.info("Reformatting DAC Sweep data")
        self.progress.append(
            (
                0,
                "Formatting DAC Sweep Data",
            )
        )
        average_per_sample = self._format_dac_sweep_to_sample(dac_data)

        _, out_minmax = self._linear_region(dac_data, minimum_r2, minimum_range)

        # Intercept and slope per sample
        for chan in range(self.channels):
            if self._cancel:
                return None

            if self.min_dac_val[chan] is None or self.max_dac_val[chan] is None:
                LOGGER.error(
                    f"Channel {chan} does not have a \
                linear region!"
                )
                continue
            LOGGER.info(f"Calculating slope & int. for channel {chan}")
            self.progress.append(
                (
                    int(20 + 80 * ((chan) / self.channels)),
                    f"Performing Linear Regression for Channel {chan}",
                )
            )
            sample_int, sample_s = self._linear_reg_for_multi_samples(
                voltages,
                average_per_sample[chan],
                vrange=(self.min_dac_val[chan], self.max_dac_val[chan]),
            )
            sample_intercept.append(sample_int)
            sample_slope.append(sample_s)

        if self._cancel:
            return None

        self.progress.append((100, "Linear Regression Completed"))

        lr_dict = {
            "slope": sample_slope,
            "intercept": sample_intercept,
            "linear_region": out_minmax,
        }

        return lr_dict

    def _linear_region(
        self, dac_data: dict, minimum_r2: float = 0.9998, minimum_range: int = 10
    ):
        """Detect the linear region
        Given a DAC Sweep, this will detect the linear region by taking all possible
        combination of two endpoints and calculate the linear regression and r-squared
        value. If it is above the minimum values, it will prefer the endpoints which
        give a greater range. If there are multiple pairs of endpoints with the same range,
        it will prefer the one with a higher r-squared value.

        Args:
            dac_data (dict): DAC sweep data collected frmo dac_sweep_controller
            minimum_r2 (float): The minimum R-squared value a region needs to be considered
                linear.
            minimum_range (int): The minimum number of points within the two endpoints to
                be considered a region. Used to optimize calculations of endpoints.

        Returns:
            out_r2 (list): Floats of R-squared value of the detected region
            out_minmax (list): Tuples of start and ending points of the linear region,
                given as mV.

        """
        voltages = self._dac2mv(np.array(list(dac_data.keys())))
        self.min_dac_val = [None] * self.channels
        self.max_dac_val = [None] * self.channels
        average_per_channel = [[] for _ in range(self.channels)]

        new_data_array = self._remove_event_dict_from_dac_sweep(dac_data)

        # Mean the data points so average_per_channel[Channel #][DAC #]
        average_per_channel = np.swapaxes(
            np.mean(np.mean(np.mean(new_data_array, axis=1), axis=1), axis=2), 0, 1
        )

        out_r2, out_minmax = self._scan_linear_region(
            voltages, average_per_channel, self.channels, minimum_r2, minimum_range
        )
        return out_r2, out_minmax

    def _scan_linear_region(
        self,
        voltages,
        average_per_channel,
        num_channels,
        minimum_r2,
        minimum_range,
    ):
        """Scans the dataset for the linear region.

        The average per channel is a curve made of the averages of all the data.
        The method uses grid search to check for the two end points with the lowest
        r2 score between them.

        TODO: Horrendously slow, this detection is not exact and is very slow O(n^2).
        Takes the average for a channel and tries to find the endpoints

        Args:
            voltages:
            average_per_channel:
            num_channels:
            minimum_r2:
            minimum_range:
        """
        out_r2 = [0] * num_channels
        out_minmax = [(0, 0) for _ in range(num_channels)]
        for chan in range(num_channels):
            max_range = 0
            for i in range(0, len(voltages) - minimum_range):
                for j in range(i + minimum_range, len(voltages)):
                    r2_xdata = []
                    intercept, coeff = self._single_sample_lr(
                        voltages[i:j], average_per_channel[chan][i:j]
                    )
                    r2_xdata.append(np.array(coeff) * voltages[i:j] + intercept)
                    r2 = r2_score(r2_xdata[0], average_per_channel[chan][i:j])
                    # Checks to see if the current endpoints are better than the current best
                    if r2 >= minimum_r2:
                        # We want as big of a range as possible
                        if (j - i) > max_range:
                            max_range = j - i
                            out_r2[chan] = r2
                            out_minmax[chan] = (int(voltages[i]), int(voltages[j]))
                        # If its the same size, we want to keep the better r2 value
                        elif (j - i) == max_range:
                            if r2 > out_r2[chan]:
                                out_r2[chan] = r2
                                out_minmax[chan] = (int(voltages[i]), int(voltages[j]))
            # Did not find a linear region
            if out_r2[chan] == 0:
                LOGGER.error(f"Did not detect a linear region in channel {chan}")
            else:
                self.min_dac_val[chan] = out_minmax[chan][0]
                self.max_dac_val[chan] = out_minmax[chan][1]
        return out_r2, out_minmax

    def _remove_event_dict_from_dac_sweep(self, dac_data):
        """Removes the event dictionary from a dac sweep, replacing
        each event dict with event['data'], and reformats array to
        recognize the newly added dimensions from the data array."""
        # TODO: Change DAC sweep data format to be dict major first,
        # So this function will be obsolete. Ex. dacdata['data'][0][0][0][0][0]
        new_data = np.array(list(dac_data.values()))
        new_data_array = np.empty(
            new_data.shape + np.array(new_data[0][0][0]["data"]).shape
        )
        for i in range(len(new_data)):
            for j in range(len(new_data[i])):
                for k in range(len(new_data[i][j])):
                    # events are dicts, remove dicts to process as numpy array
                    new_data[i][j][k] = new_data[i][j][k]["data"]
                    # Make numpy recognize ['data'] array as new dimensions
                    new_data_array[i][j][k] = np.reshape(
                        np.concatenate(new_data[i][j][k]), new_data_array[i][j][k].shape
                    )
        return new_data_array

    def _format_dac_sweep_to_sample(self, indata):
        """
        Changes the data format of dac sweep data to be sample major,
        dac val minor, for a specific channel. (outdata[sample #][dac_val])
        Args:
            indata (dict): DAC Sweep Data

        Returns:
            outdata (nparray): Data sample major, dac val minor
        """

        linearity_data = self._create_linearity_data(indata)
        averaged_data = []
        # average the 10 acquisitions for each capture
        for dac_val, dac_data in linearity_data.items():
            averaged_data.append(np.mean(dac_data, axis=0))

        sample_data = []
        for item in averaged_data:
            sample_data.append(item)
        # sample_data is in format sample_data[dac val][channel #][sample #]
        # fixed_sample_data: fixed_sample_data[channel #][sample #][dac val]
        fixed_sample_data = np.moveaxis(np.array(sample_data), 0, 2)
        return fixed_sample_data

    def _linear_reg_for_multi_samples(self, voltages, sample_data, vrange=(600, 1000)):
        """Run linear regression for each samples in the data.

        voltages is the xaxis and sample_data a multidim y_axis.
        By running it for each sample two arrays with 1.intercepts and 3. coeffs are returned.

        Args:
            voltages: voltages for the xaxis
            sample_data: matrix of datapoints for each samples along the voltage axis
            vrange: (min, max) voltages for the linear region in the data.

        Returns:
            list of intercepts, list of coefficients (`slopes`)
        """
        minv = self._find_nearest(voltages, vrange[0])
        maxv = self._find_nearest(voltages, vrange[1])
        intercepts = np.array([])
        slopes = np.array([])
        for sample in sample_data:
            intercept, slope = self._single_sample_lr(
                voltages[minv:maxv], sample[minv:maxv]
            )
            intercepts = np.append(intercepts, intercept)
            slopes = np.append(slopes, slope)

        return intercepts, slopes

    def _single_sample_lr(self, xdata, ydata):
        """Perform lienar regression of the provided samples.

        length of xdata and ydata must match or the linear regression will fail.

        Args:
            xdata: list of values along the xaxis
            ydata: list of values for the yaxis

        Returns:

        """
        if type(xdata) is list:
            xdata = np.array(xdata)
        reshaped_x = xdata.reshape(-1, 1)
        regressor = LinearRegression()
        regressor.fit(reshaped_x, ydata)

        return regressor.intercept_, regressor.coef_

    def _find_nearest(self, array, value):
        """Find the index of the sample in `array` with the value nearest `value`

        Subtract the value from the array and return the index with the smallest difference.

        Args:
            array: list of values
            value: target value to reference with `array` values

        Returns:
            index of the closest sample
        """
        array = np.asarray(array)
        idx = (
            np.abs(array - value)
        ).argmin()  # Returns the indices of the minimum values along an axis.
        return idx

    def _create_linearity_data(self, indata):
        """Create linearity data by reshaping the pedestals blocks into a full structure.

        TODO: This function can probably be replaced with a simple reshape.

        Args:
            indata: {dac_val, validated_blocks}

        Returns:
            {dac_val, list_combined_blocks}
        """
        outdata = {}
        for dac_val, data in indata.items():

            odata = self._combine_data(data)
            outdata[dac_val] = odata

        return outdata

    def _combine_data(self, data):
        """Combine data from blocks to a full data structure

        Each block contains X windows and Y events.
        Assume all blocks of equal length.
        Output should be Y events with X*blocks windows

        Args:
            data(list): list of list
            board (naludaq.board): Board object used for params
        Returns:
            list of data

        """
        if len(data) <= 1:
            raise ValueError("Data doesn't contain multiple blocks")
        # Variables
        windows = self.board.params["windows"]
        block_size = self.board.params["pedestals_blocks"]
        samples = self.board.params["samples"]
        block_len = len(data[0])  # Assume all blocks same len.
        # data structures
        output_data = []

        for evt_num in range(block_len):
            oevt_data = [[] for _ in range(self.channels)]
            for block in data:
                evt = block[evt_num]

                for chan_num, chan in enumerate(evt["data"]):
                    oevt_data[chan_num].extend(chan[0 : block_size * samples])

            # 254 windows, not 256, toss last two that wraps around.
            # TODO: This is dangerous, can we assume this is always true? depends on the configuration?
            for chan in range(self.channels):
                oevt_data[chan] = oevt_data[chan][0 : windows * samples]

            output_data.append(oevt_data)

        return output_data

    def _dac2mv(self, dac_val):
        return (
            dac_val
            * self.board.params["ext_dac"]["max_mv"]
            / (self.board.params["ext_dac"]["max_counts"])
        )



import logging

from .default import BoardController

LOGGER = logging.getLogger("naludaq.board_controller_aodsoc")


class BoardControllerAodsoc(BoardController):
    """Board controller for AODSOC boards.

    Has a special version of the `read_scalers` function to handle reading from chips
    individually.
    """

    def read_scalar(self, channel: int) -> int:
        """Read the scalar for the given channel"""
        relative_channel = self._get_relative_channel(channel)
        selected_chips = self._get_selected_chips()
        self._select_chip_by_channel(channel)
        result = super().read_scalar(relative_channel)
        self._select_chips(selected_chips)
        return result

    def _get_relative_channel(self, channel: int) -> int:
        """Get the channel relative to the chip it belongs to."""
        num_chips = self._get_num_chips()
        channels_per_chip = self.board.channels // num_chips
        return channel % channels_per_chip

    def _select_chip_by_channel(self, channel: int):
        """Enables only the chip that the given channel belongs to."""
        num_chips = self._get_num_chips()
        channels_per_chip = self.board.channels // num_chips
        chip = channel // channels_per_chip
        self._select_chips([chip])

    def _select_chips(self, chips: list[int]):
        """Select the chips enabled for RX/TX.
        Useful for reading/writing registers for a single chip.

        Args:
            chips (list[int]): list of chips that are enabled.
        """
        en = sum([1 << x for x in chips])
        self._write_control_register("ard_tx_en", en)
        self._write_control_register("ard_rx_en", en)

    def _get_selected_chips(self) -> list[int]:
        """Get the chips which are currently enabled for TX."""
        tx_en = self.board.registers["control_registers"]["ard_tx_en"]["value"]
        chips = self._get_num_chips()
        selected_chips = [(tx_en >> x) & 1 == 1 for x in range(chips)]
        return selected_chips

    def _get_num_chips(self):
        """Get the number of chips on the board"""
        return self.board.params.get("num_chips", 2)



"""Class containing all functions to control the boards.

All functions to communicate with the Nalu boards are
collected here. The Facade naludaq is intended to simplify
the interface for general use.
"""
import logging

from .aodsoc import BoardControllerAodsoc
from .default import BoardController
from .hdsoc import HDSoCBoardController
from .oleas import BoardControllerOleas
from .trbhm import TrbhmBoardController
from .udc import UDCBoardController
from .upac import UpacBoardController
from .upac96 import UPAC96BoardController

LOGGER = logging.getLogger(__name__)  # pylint: disable=invalid-name


def get_board_controller(board):
    """Get the correct board controller based on the board model.

    Args:
        board (Board): the board object.

    Returns:
        The appropriate BoardController for the given board.
    """
    return {
        # 'aodsoc_aods': BoardControllerAodsoc,
        "aodsoc_aods": BoardControllerOleas,
        "aodsoc_asoc": BoardControllerAodsoc,
        "hdsocv1": HDSoCBoardController,
        "hdsocv1_evalr1": HDSoCBoardController,
        "hdsocv1_evalr2": HDSoCBoardController,
        "trbhm": TrbhmBoardController,
        "udc16": UDCBoardController,
        "upac32": UpacBoardController,
        "upaci": UpacBoardController,
        "upac96": UPAC96BoardController,
        "zdigitizer": UpacBoardController,
    }.get(board.model, BoardController)(board)



"""Tool for automatically sending software triggers over time.

See the class docstring for more details.

Author:
    Mitchell Matsumori-Kelly <mitchell@naluscientific.com>
"""
import logging
from typing import Callable

from naludaq.controllers import get_board_controller
from naludaq.helpers import type_name
from naludaq.helpers.exceptions import AutoTriggerError
from naludaq.models.acquisition import AcquisitionLike
from naludaq.tools.autoaction import AutoAction

logger = logging.getLogger("naludaq.tools.autotrigger_default")


class DefaultAutoTrigger(AutoAction):
    def __init__(self, board, interval: "float | Callable") -> None:
        """Default auto trigger module. Used to send software triggers at a consistent interval.

        There are three main ways this class can be used in:
        1. Through the `start()` method. This will start sending software triggers in a new thread, and must
        be stopped with the `stop()` method. This is best suited for situations when the exact
        number of triggers needed is unclear or unknown.
        2. Through the `send_triggers()` method. This will send an exact number of triggers. This is
        best suited for situations in which the exact amount of data required is known.
        3. Through the `send_triggers_until_amount_read()` method. This will send triggers until some number
        of additional events have been added to a buffer.

        This class is also a context manager. Use in this manner is equivalent to `start()` on entering
        the context and `stop()` upon leaving.
        """
        super().__init__(self._send_trigger, interval=interval)
        self._board = board

        # triggers can fire quickly, so keeping just one instance is faster
        self._board_controller = get_board_controller(board)
        self._trigger_count = 0

    @property
    def board(self):
        """Get the board object"""
        return self._board

    def send_triggers(self, num_triggers: int):
        """Send a given number of software triggers at the set interval.

        This function is blocking, and will only return once all triggers have been sent.

        Args:
            num_triggers (int): the number of triggers to send. Must be a positive integer.

        Raises:
            AutoTriggerError: if the autotrigger has been started in a different thread
                and is not yet stopped, or if the software triggers cannot be sent
        """
        if not isinstance(num_triggers, int):
            raise TypeError("Argument must be a positive int")
        if num_triggers <= 0:
            raise ValueError("Argument must be a positive int")
        if self._running:
            raise AutoTriggerError(
                "Autotrigger is already running from a different thread"
            )

        logger.info("Sending %s triggers...", num_triggers)
        self._trigger_count = 0
        self.stop_condition = (
            lambda: self._trigger_count >= num_triggers
        )  # >= for safety
        self.start(blocking=True)

    def send_triggers_until_amount_read(
        self,
        buffer: AcquisitionLike,
        amount: int,
        attempts: int = 5,
    ):
        """Send triggers until a buffer increases in size by a given amount. This function
        is blocking.

        This method is mainly useful for collecting a minimum number of events when used in
        combination with the output buffer from a daq.

        The initial contents of the buffer is ignored when determining how much data has been
        read back. This means you can call this method multiple times with the same output buffer
        and you will (ideally) get `amount` new events added each time.

        Args:
            buffer (AcquisitionLike): the buffer to monitor the size of. Should be the output buffer
                from a DAQ, otherwise using this method is pointless.
            amount (int): the minimum buffer size (number of events) desired.
            attempts (int): maximum number of times to continue sending software triggers when

        Raises:
            AutoTriggerError: if there was a problem sending software triggers, or if not enough data
                was collected in the given number of attempts.
        """
        self._validate_send_triggers_until_amount_read_args(buffer, amount, attempts)

        initial_amount = len(buffer)
        while len(buffer) - initial_amount < amount:
            buff_size_before_trigger = len(buffer)
            additional_amount_needed = amount - (len(buffer) - initial_amount)

            # Will raise an error if ZERO additional events have been read by the end of the loop
            for _ in range(attempts):
                try:
                    self.send_triggers(additional_amount_needed)
                except Exception as e:
                    raise AutoTriggerError("Failed to send software triggers") from e

                additional_amount_received = len(buffer) - buff_size_before_trigger
                if additional_amount_received != additional_amount_needed:
                    logger.warning(
                        "Sent %s triggers but received %s events",
                        additional_amount_needed,
                        additional_amount_received,
                    )
                if additional_amount_received > 0:
                    break
            else:
                raise AutoTriggerError(
                    f"Failed to collect the requested amount of data ({len(buffer)-initial_amount} / {amount})"
                )

    def _send_trigger(self):
        """Send a software trigger to the board."""
        self._trigger_count += 1
        try:
            self._board_controller.toggle_trigger()
        except Exception as e:
            raise AutoTriggerError("Failed to send software trigger") from e

    # ================================= VALIDATION =================================
    def _validate_send_triggers_until_amount_read_args(
        self,
        buffer: AcquisitionLike,
        amount: int,
        attempts: int,
    ):
        """Validate the arguments passed to `send_triggers_until_amount_read_args()`"""
        if not isinstance(buffer, AcquisitionLike):
            raise TypeError(f"Buffer must be Acquisition-like, not {type_name(buffer)}")
        if not isinstance(amount, int):
            raise TypeError(f"Amount must be an int, not {type_name(amount)}")
        if not isinstance(attempts, int):
            raise TypeError(f"Attempts must be positive, not {type_name(attempts)}")
        if amount <= 0:
            raise ValueError(f"Amount must be positive (got {amount})")
        if attempts <= 0:
            raise ValueError(f"Attempts must be positive (got {attempts})")



import logging
import socket
import time

from naludaq.board.connections._UART import UART

logger = logging.getLogger(__name__)

# Import packet size frmo mockboard to standardize packets
try:
    from mockboard.params import MAX_PACKET_SIZE
except:
    MAX_PACKET_SIZE = 1024


class MockUART(UART):
    def __init__(self, conn_info: dict):
        """Class for mocking a UART connection over a UDP socket.

        Args:
            conn_info (dict): the connection configuration
        """
        super(MockUART, self).__init__(conn_info)

        self._user_addr = (conn_info["ip"], conn_info["user_port"])
        self._board_addr = (conn_info["ip"], conn_info["board_port"])

        self.ser: _MockSerial = None

    @property
    def board_addr(self) -> tuple:
        """Address used to communicate with the mock board."""
        return self._board_addr

    @property
    def user_addr(self) -> tuple:
        """Address the naludaq-side socket binds to."""
        return self._user_addr

    def open(self) -> "_MockSerial":
        """Opens the connection."""
        if self.ser:
            if not self.ser.is_open:
                self.ser.open()
            return self.ser

        self.ser = _MockSerial(self._user_addr, self.board_addr)
        return self.ser

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, value):
        self._port = value


class _MockSerial:
    def __init__(self, user_addr, board_addr):
        """Mock serial connection, implements the necessary
        functionality from pyserial but through sockets.

        Args:
            user_addr (tuple): user-side address the socket binds to.
            board_addr (tuple): address of the mock board
        """
        self.user_addr = user_addr
        self.board_addr = board_addr
        self.socket: socket.socket = None

        self.rtscts = False
        self.baudrate = 100_000
        self.input_buffer = bytearray()
        self.timeout = 0.01

        self.open()

    def __del__(self):
        self.close()

    def open(self):
        """Opens the connection."""
        if not self.socket:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.socket.setblocking(False)
            self.socket.bind(self.user_addr)

    def close(self):
        """Closes the connection."""
        try:
            self.socket.close()
        except:
            pass
        self.socket = None

    @property
    def is_open(self) -> bool:
        """Indicates whether the connection is open."""
        return self.socket is not None

    def read(self, size=1):
        """Reads at most the given number of bytes from the socket.
        May read less if the operation times out.

        Args:
            size (int): number of bytes to read.

        Returns:
            A bytearray of the data read.
        """
        start_time = time.perf_counter()
        while (
            len(self.input_buffer) < size
            and time.perf_counter() - start_time < self.timeout
        ):

            try:
                buff, _ = self.socket.recvfrom(MAX_PACKET_SIZE)
                self.input_buffer += buff
            except Exception as e:
                pass
        result = self.input_buffer[:size]
        self.input_buffer = self.input_buffer[size:]
        return result

    def read_all(self):
        """Reads all packets from the socket, until it is empty.

        Returns:
            result (bytearray): The input buffer of MockUART, populated
                with packets read from socket
        """
        while True:
            try:
                buff, _ = self.socket.recvfrom(MAX_PACKET_SIZE)
                self.input_buffer += buff
            except Exception as e:
                break

        result = self.input_buffer
        self.input_buffer = bytearray()
        return result

    def write(self, data: bytes):
        """Sends the data over the socket, to the destination port.

        Args:
            data (bytes): The bytes to send over socket
        """
        try:
            self.socket.sendto(data, self.board_addr)
        except:
            raise IOError("Could not send data over socket.")

    @property
    def in_waiting(self):
        """Reads waiting packets in the socket connection, and
        appends it to the input buffer. Returns the number
        of bytes in wait.

        Returns:
            length (int): Length of the input buffer byte array.
        """
        self.input_buffer = self.read_all()
        return len(self.input_buffer)

    def reset_input_buffer(self):
        """Empties out the packets waiting in socket, as well as
        the input buffer.
        """
        self.read_all()

    def reset_output_buffer(self):
        pass



import logging
import time
from typing import List

import numpy as np

from naludaq.communication import sendI2cCommand
from naludaq.controllers.external_dac.i2c_dac import I2CDACController
from naludaq.helpers.exceptions import InvalidBoardModelError

LOGGER = logging.getLogger("naludaq.ext_dac.dac7578")


class DACControllerDAC7578(I2CDACController):
    """DAC controller for boards using the DAC7578 DAC."""

    def __init__(self, board, send_delay=0.001) -> None:
        """Controller for the DAC7578 DAC.

        Args:
            board (Board): the board object.
            send_delay (float): the time to wait between sequential I2C
                writes in seconds.
        """
        super().__init__(board, send_delay=send_delay)
        self._write_cmd = 0b0011

        if board.params["ext_dac"]["chip"] != "dac7578":
            raise InvalidBoardModelError(
                "The given board cannot be used with this DAC controller"
            )

    def _write_dacs(self, channels: List[int]):
        """Writes the values to the DAC for each board channel given.

        The values written come from the board object.

        Args:
            channels (List[int]): list of channels to set the
                DACs for.

        Raises:
            TypeError if the given channels are of the wrong type
            ValueError if any of the internal DAC values on the board
                are out of bounds.
            ConnectionError if the I2C commands could not be sent
        """
        try:
            self._validate_channels(channels)
            self._validate_internal_dac_values()
        except (TypeError, ValueError):
            raise

        if (
            set(channels) == set((range(self.board.channels)))
            and len(set(self.board.dac_values)) == 1
        ):
            self._broadcast_to_all_channels()
            return

        self._write_to_channels(channels)

    def _write_to_channels(self, channels: List[int]) -> None:
        """Writes the external dac values per channel in channels"""

        rw = 0
        LOGGER.info(
            f"Setting channels {channels} to: {[self.board.dac_values[channel] for channel in channels]}"
        )

        for channel in channels:
            value = self.board.dac_values[channel]
            addr = self._address_mapping[channel]
            words = self._get_write_words(channel, value)

            LOGGER.debug(
                "Sent DAC Command - ADDR: %s, WORDS: %s", (addr << 1) | rw, words
            )
            if not sendI2cCommand(self.board, (addr << 1) | rw, words):
                raise ConnectionError("Failed to write I2C DAC")
            time.sleep(self._send_delay)

    def _broadcast_to_all_channels(self):
        """Broadcast  to every channel on EVERY ext. DAC chip."""
        rw = 0
        unique_addrs = self._get_unique_addresses()

        for addr in unique_addrs:
            addr = int(addr)
            value = self.board.dac_values[0]
            words = self._get_broadcast_words(value)
            LOGGER.info("Broadcasting DAC value: %s to DAC %s", value, addr)
            LOGGER.debug("Sent DAC Command: %s", words)
            if not sendI2cCommand(self.board, (addr << 1) | rw, words):
                raise ConnectionError("Failed to write I2C DAC")
            time.sleep(self._send_delay)

    def _get_write_words(self, channel: int, value: int) -> List[str]:
        """Get words to write to set the DAC value for a channel.

        Args:
            channel (int): the channel to set the DAC value for.
            value (int): the DAC value.

        Returns:
            A list of bytes to send over the I2C bus.
        """
        return [
            f"{self._write_cmd:04b}{self._channel_mapping[channel]:04b}",
            f"{(value >> 4 & 0xFF):08b}",
            f"{(value << 4 & 0xFF):08b}",
        ]

    def _get_broadcast_words(self, value: int) -> str:

        return [
            f"{self._write_cmd:04b}{0b1111:04b}",
            f"{(value >> 4 & 0xFF):08b}",
            f"{(value << 4 & 0xFF):08b}",
        ]

    def _get_unique_addresses(self):
        return list(np.unique(list(self._address_mapping.values())))



"""Control Register
=================

Control registers change functinality of the FPGA.

The control registers are FPGA specific values,
by changing these, the FPGA changes it's operation.

They are stored in two places, in the board object and
in the physical hardware. It's possible to change them independently.

Command structure:
------------------
Updating a digital command uses a format 'CCAAVVVV'

- CC (8-bit): tells the FPGA command parser to expect a control register command.
- AA (8-bit): address
- VVVV (16-bit): value

Will be interpereted by the FPGA as hex. it's 32-bit = 8 hex values.

Examples:
Read the values back from the board
ctrl_reg_vals = read_control_registers(board)

If there is a missmatch the board-objects value can be overwritten:
read_control_registers(board, overwrite=True)

Examples:
----------

.. code-block: python

    dc = DigitalRegisters(board)
    dc.read("regname")
    dc.write("regname", intvalue)

"""
from logging import getLogger

from naludaq.communication import registers

LOGGER = getLogger(__name__)


class ControlRegisters(registers.ReadRegisters, registers.WriteRegisters):
    """Control Registers

    Attributes:
        board:
    """

    def __init__(self, board):
        super().__init__(board, "control_registers")

        self._write_cmd = 0xAF
        self._read_cmd = 0xAD
        self.width_cmd = 2
        self.width_addr = 2  # hex char
        self.width_val = 4  # hex char



import time
from collections import deque

from naludaq.board import Board
from naludaq.communication import AnalogRegisters, ControlRegisters, DigitalRegisters
from naludaq.controllers import get_board_controller, get_readout_controller

# from naludaq.tools.pedestals.pedestals_controller import sleep_calc, increase_margin


class DataCollector:
    def __init__(self, daq, board: Board):
        """Wrapper for the standard readout procedure for most Nalu Scientific ASICs

        Takes all of the configuration settings for readout and updates the settings to the chip
        when readout is commenced. The Data Collector is very specific with its settings, and will
        only take in the correctly formatted settings.

        Args:
            daq (BenDaq, LightDaq): The Data acquisition object that the DataCollector will use to
                interface with the chip
            board (Board): The Board object that has the proper parameters and connection to the
                physical board
        """

        self.board = board
        self.daq = daq
        self.default_margin = 1.1
        self.margin = self.default_margin
        self.sleeptime = sleep_calc(board, self.margin)

    @property
    def board(self):
        return self._board

    @property
    def daq(self):
        return self._daq

    @board.setter
    def board(self, board):
        self._board = board

    @daq.setter
    def daq(self, daq):
        self._daq = daq

    def get_data(
        self,
        amount: int,
        trig_type: str,
        lb_type: str,
        windows: int,
        lookback: int,
        write_after_trig: int,
        channels: list,
        timeout: int = 5,
    ) -> deque:
        """Returns an exact amount of pre-processed events!

        The return is sizelimited to `amount` by a circular buffer, it'll overwrite the oldest data
        if the board returns too much data. This is by design and can't be changed.


        ## Readout Settings Explanation:
            `trig_type` (trigger_type):
                'imm' (immediate): digitize whatever instantly as fast as possible.
                'ext' (external): use the external trigger.
                'self' (self): trigger on input RF signal. Uses trigger values
            `lb_type` (lookback type):
                'forced' (forced): always reads out same set of windows, start window set by lookback
                'trigrel' (trigger relatve): relative to the trigger.
                'roi' (region of interest): complicated...

        Args:
            `amount` (int): desired amount of events, cannot be less than 1
            `trig_type` (string): How the trigger runs for the acquisition
            `lb_type` (string): The type of lookback for the acquisition
            `windows` (int): amount of windows to read. Value must be a positive integer
            `lookback` (int): how many windows the readout header goes back
            `write_after_trig` (int): how many windows are read out after the trigger signal is recieved
            `channels` (List): list of channels to be activated when reading out data
            `timeout` (int): timeout in retries per event.

        Returns:
            A deque with the events, raw or processed depending on the daq.

        Raises:
            `ValueError` if amount is < 1.
            `TimeoutError` if no data is captured in `timeout` retries.
        """

        rdt_controller = get_readout_controller(self.board)
        readout_settings = {
            "trig": trig_type,
            "lb": lb_type,
            "acq": "pedsub",  # Hardcoded because all simple readouts use this mode
            "ped": "zero",  # Hardcoded because all simple readouts use this mode
            "readoutEn": True,  # This should always be on since we always want ethernet readout
            "singleEv": True,  # Always on, make sure `amount` counter is used by FW.
        }

        window_settings = {
            "windows": windows,
            "lookback": lookback,
            "write_after_trig": write_after_trig,
        }

        self._validate_readout_settings(readout_settings)
        self._validate_window_settings(window_settings)
        self._validate_channel_settings(channels)
        if amount < 1:
            raise ValueError(f"amount cannot be less than 1")

        output_buffer = deque(maxlen=amount)
        self.daq.output_buffer = output_buffer
        desired_amount = amount
        reset_count = 0
        self.margin = self.default_margin
        rdt_controller.set_read_window(**window_settings)
        rdt_controller.set_readout_channels(channels)
        while len(output_buffer) < amount:
            desired_amount = amount - len(output_buffer)
            if desired_amount <= 0:
                break

            self.sleeptime = sleep_calc(self.board, self.margin)

            rdt_controller.number_events_to_read(desired_amount)
            self.daq.start_capture()
            get_board_controller(self.board).start_readout(**readout_settings)

            finished = self._wait_for_data(output_buffer, desired_amount)

            if not finished:
                if reset_count >= timeout:
                    raise TimeoutError(
                        f"Something wrong with setup/hardware, "
                        f"reset {timeout} times still no data."
                    )
                reset_count += 1
                self.margin = increase_margin(self.margin)
                self.sleeptime = sleep_calc(self.board, self.margin)

            self.daq.stop_capture()
            get_board_controller(self.board).stop_readout()

        return output_buffer

    def _wait_for_data(self, output_buffer, desired_amount):
        """Wait for the desired amount of data to arrive.

        Returns in these cases:
        - if desired amout is reached
        - if no data is received
        - if too much data is received

        Args:
            output_buffer (deque): storage for data we wait for.
            desired_amount(int): The amount of events we wait for.

        Returns:
            True if the correct amount of data is captured, else False
        """
        result = True
        prev_amount = -1
        # wait for readout to finish
        while len(output_buffer) != prev_amount:
            stime = self.sleeptime * (desired_amount - len(output_buffer))
            if output_buffer:
                prev_amount = len(output_buffer)
            try:  # If output_buffer is larger than desired, stime < 0.
                time.sleep(stime)
            except ValueError:
                pass
            if len(output_buffer) > desired_amount:
                # Output buffer can currently never be bigger since it's size limited.
                # Keep edge case in case someone changes the limit.
                break

        if len(output_buffer) == 0:
            result = False

        return result

    def parameter_sweep(self, parameter, numevts, param_min, param_max, param_step=1):
        """
        Iterates through values in a specified register, taking data for every value in the sweep.

        Register type (analog, digital, or control) is automatically identified.

        Args:
            parameter (str): name of register to sweep through
            numevts (int): number of events to take at each parameter value
            param_min (int): minimum parameter value
            param_max (int): maximum parameter value
            param_step (int): sweep step value

        Returns:
            List of data taken coupled with the parameter value used during that acquisition

        Raises:
            ValueError when parameter does not match existing registers
        """

        data_manifest = []

        register_controller = self._register_type_selector(parameter)

        for param_value in range(param_min, param_max, param_step):
            register_controller.write(parameter, param_value)

            data = self.get_data(numevts)
            data_row = [param_value, data]
            data_manifest.append(data_row)

        return data_manifest

    def _register_type_selector(self, parameter):
        if parameter in AnalogRegisters(self.board).registers.keys():
            reg_com = AnalogRegisters(self.board)
        elif parameter in DigitalRegisters(self.board).registers.keys():
            reg_com = DigitalRegisters(self.board)
        elif parameter in ControlRegisters(self.board).registers.keys():
            reg_com = ControlRegisters(self.board)
        else:
            raise ValueError(f"{parameter} is not a valid register name")

        return reg_com

    @staticmethod
    def _validate_readout_settings(readout_settings: dict):

        valid_readout_settings = {
            "trig": ["i", "s", "e"],
            "lb": ["t", "f", "r"],
            "acq": ["p", "r"],
            "ped": ["z", "c", "r"],
            "readoutEn": [True, False],
            "singleEv": [True, False],
        }

        # Validate values
        for key in readout_settings.keys():
            if key not in ["readoutEn", "singleEv"]:
                if readout_settings[key][0] not in valid_readout_settings[key]:
                    raise KeyError(
                        f"{readout_settings[key]} is not a valid value for {key}. Valid \
                        values are f{valid_readout_settings[key]}"
                    )
            else:
                if readout_settings[key] not in valid_readout_settings[key]:
                    raise KeyError(
                        f"{readout_settings[key]} is not a valid value for {key}. Valid \
                        values are f{valid_readout_settings[key]}"
                    )

    @staticmethod
    def _validate_window_settings(window_settings: dict):

        # Validate values
        for key in window_settings.keys():
            if not isinstance(window_settings[key], int):
                raise KeyError(
                    f"{key} does not have an integer value. All window settings should \
                    be integer values"
                )

    def _validate_channel_settings(self, channels: list):

        valid_channels = set(range(self.board.params["channels"]))

        if not set(channels).issubset(valid_channels):
            raise TypeError(
                f"{channels} is not in range of valid channels for this board model. \
                Valid range is from 0 to {self.board.params['channels']}"
            )


def increase_margin(margin: float, margin_inc: float = 1.5, max_margin: float = 5):
    """Increases the margin, using an exponential standoff"""
    margin *= margin_inc
    margin = min(margin, max_margin)
    return margin


def sleep_calc(board, margin: float = 2.0) -> float:
    """Compute the `sleeptime` for the board.

    Using the `board.params` to gather blocksize, baudrate and readout
    parameters compute the minumum sleep time.

    The `margin` can be dynamically altered as a standoff in case the calc
    doesn't match real transfer rate.
    By increasing `margin` it's possible to increase `sleeptime` in case of
    real-world differences between calculated baudrate and real. Or if
    the return data is longer/padded.

    Args:
        board(naludaq.board)
        margin(float): Multiplier to increase sleeptime due to real-world
            scenarios.

    Returns:
        sleeptime in seconds as a float

    """
    baudrate = board.connection_info.get("speed", 115200)
    transferrate = baudrate // 8
    windows = board.params.get("pedestals_blocks", 16)
    channels = board.params.get("channels", 4)
    samples = board.params.get("samples", 64)
    data = windows * channels * samples * margin * 2  # 16-bits per value
    sleeptime = data / transferrate

    return sleeptime



"""Trigger controller.

The trigger controller is an interface to access and operate the
trigger functionality in the Nalu hardware.
The trigger is active when the chip is in the trigger mode,
by setting BoardController(board).start_readout(*args) where trig
is set to "ext" or "imm".

The trigger is operated by two variables, offset and value.
The value is the ADC threshold value of the trigger.
The offset is an offset from the 0, by setting offset it's possible to
fine-tune the trigger since the needed value is now smaller.

Disable the trigger by setting the value to 0.

Example:
    trigger = TriggerController(board)

    trig_vals = [1000, 0, 1000, 0]
    trigger.trigger_values = trig_vals

    trigger.write_triggers()

AUTHOR:
Marcus Luck <marcus@naluscientific.com>

"""
import logging
from typing import List

from naludaq.communication import AnalogRegisters, ControlRegisters, DigitalRegisters
from naludaq.controllers.controller import Controller

LOGGER = logging.getLogger(__name__)  # pylint ignore=invalid-name


class TriggerController(Controller):
    """Primary tool to setup the triggers for an aquisition.

    Attributes:
        board: The board to update the triggers on.
        trigger_offsets: Offsets as a list, one val per channel.
        trigger_values: The trigger values as a list, one val per channel.
        baseband_attempts: Times to run threshold scan to establish trigger baseband.

    Functions:
        write_triggers: writes all the trigger values to the physical hardware.
    """

    def __init__(self, board):
        super(TriggerController, self).__init__(board)
        # self.board = board
        self._wait = board.params["wait"]
        self._wbias = board.params["wbias"]
        self._num_trig_chans = board.params["channels"]
        self.baseband_attempts = 2
        self.bank = 1

    @property
    def trigger_offsets(self):
        """Set voltage offset for the trigger.

        Set the trigger offset values, can set one value for all channels or
        a specific value for each channel.

        Updates the value on the hardware if connected.

        Args:
            offset(int): voltage offset for all channels.
            offset(list): voltage offset for the individual channels.

        Returns:
            List of current voltage offsets.

        Raises:
            TypeError if voltage is not a list of length num_chans or a single number.
        """
        return self.board.trigger_offsets

    @trigger_offsets.setter
    def trigger_offsets(self, offset) -> None:
        if isinstance(offset, (int, float)):
            self.board.trigger_offsets = [offset] * self._num_trig_chans

        elif isinstance(offset, list) and len(offset) == self._num_trig_chans:
            self.board.trigger_offsets = offset

        else:
            raise TypeError("Voltage offset must be a single number or a list.")

        self.write_triggers()

    @property
    def trigger_values(self) -> List[int]:
        """Trigger threshold for the board.

        Set the trigger threshold values, can set one value for all channels or
        a specific value for each channel.

        Updates the value on the hardware if connected.

        Args:
            trigger_val (list or int): Threshold values for triggers per channel.
                Can take an single number for all channels or one value per channel.

        Returns:
            List of current values for the triggers.

        Raises:
            TypeError if not a List[int].
            ValueError if the list is the wrong length
        """
        return self.board.trigger_values

    @trigger_values.setter
    def trigger_values(self, trigger_val: List[int]) -> None:
        if isinstance(trigger_val, int):
            trigger_val = [trigger_val for _ in range(self._num_trig_chans)]
        try:
            self.board.trigger_values = trigger_val
        except:
            raise
        self.write_triggers()

    def write_triggers(self):
        """Write the trigger values to the hardware.

        This writes the current set triggers to the hardware.
        By setting the trigger_offsets or the trigger_values
        this function is called automatically and the hardware is updated.
        """
        self._set_trigger_thresholds()
        self._set_wbiases()
        self._set_trigger_offsets()

        return True

    def _set_trigger_thresholds(self):
        """Sets the trigger values and send to hardware.

        Take the trigger values from the board
        """
        trigger_values = self.board.trigger_values
        channels = self.board.params["channels"]

        reg = "trigger_threshold_{:02}"
        if self.board.params["model"] == "siread":
            reg = "thresh_{:02}"

        for chan in range(channels):
            register = reg.format(chan)
            tval = trigger_values[chan]
            self._write_analog_register(register, tval)

    def _set_wbiases(self):
        channels = self.board.params["channels"]
        wbias = [0] * (channels // self.bank)
        for index in range(channels // self.bank):
            if any(
                x > 0
                for x in self.board.trigger_values[
                    index * self.bank : index * self.bank + self.bank
                ]
            ):
                wbias[index] = self._wbias

            register = "wbias_" + str(index).zfill(2)
            self._write_analog_register(register, wbias[index])

    def _set_trigger_offsets(self):

        vofs = self.board.trigger_offsets

        channels = self.board.params["channels"]
        # if self.board.params['analog_values'].get('offset_00', None) is not None:
        areg = self.board.registers.get("analog_registers", {})
        if areg.get("offset_00", None) is not None:
            for index in range(channels // self.bank):
                register = "offset_" + str(index).zfill(2)
                self._write_analog_register(register, int(vofs[index]))

    def set_trigger_edge(self, raising: bool = True):
        """Set which signal edge to trigger on.

        Shift between positive going signals (raising) and negative (falling).

        Args:
            raising(bool): If true, trigger on positive going signals, else falling edge.

        Raises:
            TypeError if raises is not a bool.
        """
        if not isinstance(raising, bool):
            raise TypeError("raising must be a bool, got %s", type(raising))

        self._write_analog_register("sgn", raising)

    def _write_analog_register(self, register, value):
        """wrapper for the Analog register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        """
        AnalogRegisters(self.board).write(register, value)

    def _write_control_register(self, register, value):
        """wrapper for the Control register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        """
        ControlRegisters(self.board).write(register, value)

    def _write_digital_register(self, register, value):
        """wrapper for the Digital register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        """
        DigitalRegisters(self.board).write(register, value)



import logging
from typing import Iterator

from naludaq.backend.models.acquisition import RemoteAcquisition
from naludaq.controllers import get_readout_controller

from .aav3 import PedestalsGeneratorAardvarcv3New

LOGGER = logging.getLogger(
    "naludaq.pedestals_controller_hdsoc_new"
)  # pylint: disable=invalid-name


class PedestalsGeneratorHdsocNew(PedestalsGeneratorAardvarcv3New):
    """Pedestals generator for HDSoC + naludaq_rs.

    It's the same as the AARDVARCv3 pedestals generator, but with a a special fix
    which accounts for the first window being invalid. When the first window fix
    is enabled, the first window of every event captured is thrown away, since
    the first window is usually (always?) invalid. This will not affect the structure
    of the generated pedestals; all windows are still present in the output.
    """

    def __init__(
        self, board, num_captures: int, num_warmup_events: int, channels: list[int]
    ):
        super().__init__(board, num_captures, num_warmup_events, channels)
        self._first_win_fix = self.board.params.get("first_window_fix", True)

    @property
    def first_window_fix(self) -> bool:
        """Get/set whether the first window fix is enabled.
        The default is True unless otherwise specified in the board parameters.

        See the class documentation for more information on the first window fix.
        """
        return self._first_win_fix

    @first_window_fix.setter
    def first_window_fix(self, enabled: bool):
        if not isinstance(enabled, bool):
            raise TypeError("Type must be boolean")
        self._first_win_fix = enabled

    def _set_read_window(self, start_window: int):
        """Set the read window from the given start window.

        Reimplemented to account for the first window fix.
        If the first window fix is enabled, the actual start window is one less
        so that we can throw it away later, and the block size is one window more
        so we can still read until the end of the block.

        Args:
            start_window (int): the start window of the block (lookback)
        """
        get_readout_controller(self.board).set_read_window(
            windows=self._correct_block_size(self.block_size),
            lookback=self._correct_start_window(start_window),
            write_after_trig=self.board.params["windows"],
        )

    def _stream_valid_events(
        self, acq: RemoteAcquisition, start: int, expected_block: list[int]
    ) -> Iterator[dict]:
        """Generator for validated events. Does not yield any bad events,
        they are just skipped.

        Reimplemented to account for the first window fix.
        If the first window fix is enabled, the first window is thrown away
        from all events.

        Args:
            acq (RemoteAcquisition): the acquisition to stream from
            start (int): the start event in the acquisition
            expected_block (list[int]): the expected window labels
        """
        for event in super()._stream_valid_events(acq, start, expected_block):
            if self._first_win_fix:
                # this function operates in-place
                self._remove_first_window(event)
            yield event

    def _calculate_expected_window_labels(
        self, start_window: int, block_size: int
    ) -> list[int]:
        """Calculate the window labels we expect to see from an event
        read from the given block.

        Reimplmented to account for the first window fix.
        If the first window fix is enabled, the actual start window is one less
        so that we can throw it away later, and the block size is one window more
        so we can still read until the end of the block.

        Args:
            start_window (int): the start window of the block
            block_size (int): the block size

        Returns:
            list[int]: the expected window labels
        """
        return super()._calculate_expected_window_labels(
            self._correct_start_window(start_window),
            self._correct_block_size(block_size),
        )

    def _remove_first_window(self, event: dict):
        """Remove the first window from the event in-place.

        Args:
            event (dict): the event
        """
        samples = self.board.params.get("samples", 32)
        for li, lbl in [
            (i, lbl) for i, lbl in enumerate(event["window_labels"]) if len(lbl) > 0
        ]:
            event["window_labels"][li] = lbl[1:]
        for li, dat in [
            (i, dat) for i, dat in enumerate(event["data"]) if len(dat) > 0
        ]:
            event["data"][li] = dat[samples:]
            event["time"][li] = dat[samples:]

    def _correct_start_window(self, start_window: int) -> int:
        """Correct the given start window by subtracting one
        when the first window fix is enabled. Does nothing
        if the switch is off.
        """
        if self.first_window_fix:
            start_window -= 1
            if start_window < 0:
                start_window = self.board.params["windows"] - 1
        return start_window

    def _correct_block_size(self, block_size: int) -> int:
        """Correct the block size by adding one when the first
        window fix is enabled. Does nothing when the switch is off.
        """
        if self.first_window_fix:
            block_size = min(block_size + 1, self.board.params["windows"])
        return block_size



from naludaq.backend.context import Context
from naludaq.backend.exceptions import ContextError


class Manager:
    def __init__(self, board) -> None:
        if not hasattr(board, "context"):
            raise ContextError("A context is required")
        self._board = board

    @property
    def board(self):
        return self._board

    @property
    def context(self) -> Context:
        return self._board.context



"""
"""
from logging import getLogger

from naludaq.communication import ControlRegisters

from .default import BoardController

LOGGER = getLogger("naludaq.board_controller_trbhm")


class TrbhmBoardController(BoardController):
    """Special board controller for TRBHM."""

    def read_scalar(self, channel: int) -> int:
        """Read the scalar for the given channel"""
        relative_channel = self._get_relative_channel(channel)
        selected_chips = self._get_selected_chips()
        self._select_chip_by_channel(channel)
        result = super().read_scalar(relative_channel)
        self._select_chips(selected_chips)
        return result

    def _get_relative_channel(self, channel: int) -> int:
        """Get the channel relative to the chip it belongs to."""
        num_chips = self._get_num_chips()
        channels_per_chip = self.board.channels // num_chips
        return channel % channels_per_chip

    def _select_chip_by_channel(self, channel: int):
        """Enables only the chip that the given channel belongs to."""
        num_chips = self._get_num_chips()
        channels_per_chip = self.board.channels // num_chips
        chip = channel // channels_per_chip
        self._select_chips([chip])

    def _select_chips(self, chips: list[int]):
        """Select the chips enabled for RX/TX.
        Useful for reading/writing registers for a single chip.

        Args:
            chips (list[int]): list of chips that are enabled.
        """
        en = sum([1 << x for x in chips])
        self._write_control_register("ard_tx_en", en)
        self._write_control_register("ard_rx_en", en)

    def _get_selected_chips(self) -> list[int]:
        """Get the chips which are currently enabled for TX."""
        tx_en = self.board.registers["control_registers"]["ard_tx_en"]["value"]
        chips = self._get_num_chips()
        selected_chips = [(tx_en >> x) & 1 == 1 for x in range(chips)]
        return selected_chips

    def _get_num_chips(self):
        """Get the number of chips on the board"""
        return self.board.params.get("num_chips", 2)

    def toggle_trigger(self):
        """Toggles the ext trigger using software.

        For TRBHM the wait between separate register writes is too long, and
        toggling the trigger too slowly results in too many events coming back
        and filling the FIFO, causing malformed events. This method instead
        sends the register writes all as one string.
        """
        cr = ControlRegisters(self.board)

        wait_cmd = "AE000001"
        exttrig_high_cmd = cr.generate_write("exttrig", True)
        exttrig_low_cmd = cr.generate_write("exttrig", False)
        toggle_cmd = wait_cmd + exttrig_high_cmd + exttrig_low_cmd
        self._send_command(toggle_cmd)

    def set_loopback_enabled(self, enabled: bool):
        """Set whether serial loopback is enabled.

        Loopback can safely be disabled during most of the operations with the board.
        Loopback **must** be disabled when communicating over the serial interface.
        If serial communication with the ASIC is intended then this should run during startup and only be enabled as needed.

        Args:
            enabled (bool): True to enable loopback.

        Raises:
            TypeError if enabled is not a bool.
        """
        if not isinstance(enabled, bool):
            raise TypeError("Argument must be bool")
        OFF = "B0900002"
        ON = "B0900003"
        cmd = ON if enabled else OFF
        self._send_command(cmd)



import logging
import time
from collections import deque

import numpy as np

from naludaq.communication import control_registers, digital_registers
from naludaq.controllers import get_dac_controller
from naludaq.helpers.exceptions import BoardParameterError, PedestalsDataCaptureError
from naludaq.tools.pedestals import get_pedestals_controller

LOGGER = logging.getLogger(__name__)  # pylint: disable=invalid-name
DigitalRegisters = digital_registers.DigitalRegisters
ControlRegisters = control_registers.ControlRegisters


class DACSweepController:
    """DAC Sweep Controller allows the user to easily collect data for
    a sweep of dac values.

    The DAC Sweep Controller uses portions of the pedestal controller
    to collect data for every sample per channel per dac sweep, allowing
    the user to find the linear region and linear regression values on
    a per sample basis.

    Args:
        board (naludaq.board): Board object

    Attributes:
        num_captures_for_peds_avg: How many datapoints/samplepoint used to calculate the average.

    raises:
        BoardParameterError if ext_dac fields are not available amoung the board params.
    """

    def __init__(self, board):
        # The type of daq used depends on the connection type
        self._board = None
        self.board = board
        self.num_evt_per_dacval = 10
        self.dac_max_counts: int = self._get_max_counts()
        self._progress: list = []
        self._cancel = False
        self.ped_ctrl = None
        self._backup_dac: dict = {}

    def _get_max_counts(self) -> int:
        try:
            return self.board.params["ext_dac"]["max_counts"]
        except KeyError as e:
            raise BoardParameterError(
                f"The register file lacks necessary ext_dac fields: {e}"
            )

    @property
    def dac_max_counts(self):
        """Get/Set the max count value for the DAC.

        The true max value is determined by the board params.

        12-bit: <= 4095
        16-bit: <= 65535

        Raises:
            TypeError: if value is not an int
            ValueError: if value is outside of the boards accepted range.
        """
        return self._dac_max_counts

    @dac_max_counts.setter
    def dac_max_counts(self, value):
        if not isinstance(value, int):
            raise TypeError("dac max counts must be an int.")
        min_val = self.board.params["ext_dac"].get("min_counts", 0)
        max_val = self.board.params["ext_dac"].get("max_counts", 4095)
        if not 0 < value <= max_val:
            ValueError(
                f"dac_max_counts must be a value between {min_val} and {max_val}"
            )
        self._dac_max_counts = value

    @property
    def board(self):
        return self._board

    @board.setter
    def board(self, board):
        self._board = board

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError("Progress is stored in a list")
        self._progress = value

    def cancel(self):
        """Cancels the dac sweep as soon as possible.
        No data is generated, and the board is restored to
        its previous state.

        Can only be called from a separate thread.
        """
        self._cancel = True
        if self.ped_ctrl:
            self.ped_ctrl.cancel()

    def dac_sweep(
        self,
        step_size: int = 50,
        min_counts: int = 0,
        max_counts: int = None,
        events_per_datapt: int = 10,
        channels: list = None,
    ):
        """Perform a dac sweep with a given step size.
        IMPORTANT: TURN OFF CONNECTED FUNCTION GENERATORS BEFORE RUNNING
        Ext-dac + signal = :(

            Args:
                step_size (int): Step sizes for the ext-dac, from 0 to max_counts

            Returns:
                dac_sweep (dict): In form dac_sweep[Dac Val][Block #][Event #] or None
                    if the dac sweep was canceled.
                min_counts (int): minimum DAC value to start sweep from
                max_counts (int): maximum DAC value to stop sweep at
                events_per_datapt (int): number of events per data point
                channels (list): list of channel numbers to sweep over
        """
        self._cancel = False

        if step_size <= 0 or step_size > self.dac_max_counts:
            raise ValueError(f"Step size is out of bounds: 0-{self.dac_max_counts}")
        if channels is None:
            channels = list(range(self.board.channels))
        if max_counts is None:
            max_counts = self.board.params["ext_dac"]["max_counts"]

        self._backup_dac_values()

        output2analyze = {}

        ext_dac_values = np.arange(min_counts, max_counts + 1, step_size)
        LOGGER.info(
            f"Starting DAC Sweep from {min_counts}-{self.dac_max_counts} with step {step_size}"
        )

        for dac_idx, dac_val in enumerate(ext_dac_values):
            dac_val = int(dac_val)
            LOGGER.info(f"Currently running DAC_val: {dac_val}")
            # set dac value for all channels
            self.progress.append(
                (
                    int(98 * (dac_val - min_counts) / (max_counts - min_counts)),
                    f"Collecting DAC value {dac_idx+1}/{len(ext_dac_values)}",
                )
            )
            get_dac_controller(self.board).set_dacs(dac_val, channels)
            time.sleep(0.001)
            data2analyze = deque()
            try:
                self.ped_ctrl = get_pedestals_controller(self.board, channels=channels)
                self.ped_ctrl._capture_data_for_pedestals(events_per_datapt)
                data2analyze = self.ped_ctrl.validated_blocks
            except PedestalsDataCaptureError:
                LOGGER.error("Couldn't capture DAC data")

            # Readout data for all channels, all windows.
            output2analyze[dac_val] = data2analyze
            LOGGER.info(f"{dac_val} total data: {len(data2analyze)}")

            if self._cancel:
                break

        # Restore backups
        self._restore_dac_backup()

        # Data is probably bad if canceled, so return None
        if self._cancel:
            return None

        # save data
        try:
            self.progress.append((99, "Saving DAC Sweep"))
        except:
            LOGGER.debug("Status message couldn't append to self.progress")
        return output2analyze

    def _backup_dac_values(self):
        """Backup the old dac values during sweep"""
        self._backup_dac = (
            self.board.params.get("ext_dac", {}).get("channels", {}).copy()
        )

    def _restore_dac_backup(self):
        for chan, value in self._backup_dac.items():
            get_dac_controller(self.board).set_single_dac(chan, value)



"""Tool for automatically sending software triggers over time.

See the class docstring for more details.

Author:
    Mitchell Matsumori-Kelly <mitchell@naluscientific.com>
"""
import logging
import sched
import sys
import time
from contextlib import AbstractContextManager
from threading import Thread
from typing import Callable

from naludaq.helpers.exceptions import AutoActionError

logger = logging.getLogger("naludaq.tools.autotrigger_default")
_default_stop_condition = lambda: False


class AutoAction(AbstractContextManager):
    def __init__(
        self,
        action: Callable,
        stop_condition: Callable = _default_stop_condition,
        interval: float = 1,
    ):
        """Create a new AutoAction object.

        Args:
            action (Callable): a function (or function-like) object to call periodically
            stop_condition (Callable): a function (or function-like) object which this class
                will call to determine whether a stop condition has been met, at which point
                the class will stop calling the `action`. The signature must be: `() -> bool`,
                returning `True` to stop or `False` to continue. If not provided, the
                stop condition will never be met and the class must be stopped manually.
                NOTE: the stop condition is called *before* each action call
            interval (float): the interval at which to call the `action` in seconds.
        """
        self.action = action
        self.stop_condition = stop_condition
        self.interval = interval

        self._running = False
        self._thread = None
        self._scheduler = None
        self._exception = None

        self._time_fn = time.monotonic
        self._delay_fn = time.sleep

    @property
    def action(self):
        """Get the board object"""
        return self._action

    @action.setter
    def action(self, action: Callable):
        if not callable(action):
            raise TypeError("Action must be callable")
        self._action = action

    @property
    def stop_condition(self) -> Callable:
        """Get/set the stop condition function, used to progammatically
        determine when to stop action-ing.

        The function must take no arguments, and return a `bool`
        indicating whether to stop.
        """
        return self._stop_condition

    @stop_condition.setter
    def stop_condition(self, stop_condition: Callable):
        if not callable(stop_condition):
            raise TypeError("Stop condition must be callable")
        self._stop_condition = stop_condition

    @property
    def interval(self) -> float:
        return self._interval

    @interval.setter
    def interval(self, interval: float):
        if not isinstance(interval, (int, float)):
            raise TypeError("Interval must be numeric")
        if interval <= 0:
            raise ValueError("Interval must be positive")
        self._interval = interval

    @property
    def is_running(self) -> bool:
        """Check if currently running"""
        return self._running

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc_args):
        self.stop()

    def start(self, blocking=True):
        """Starts calling the action periodically from a new thread.

        Args:
            blocking (bool): whether to block until the stop condition is met
                or the `stop()` method is called.
        """
        if self._running:
            raise AutoActionError("Already running")

        # scheduler used to periodically call action
        self._scheduler = self._make_scheduler()
        self._schedule_action(self._time_fn())  # initial action, occurs immediately

        self._running = True
        self._thread = Thread(target=self._scheduler.run, daemon=True)
        self._thread.start()
        if blocking:
            self._thread.join()
            self._running = False
            self._reraise_last_error()

    def stop(self):
        """Tells the autoaction module to cease scheduling new actions.

        This method blocks until the scheduler is complete.
        """
        self._running = False
        if self._thread is not None:
            self._thread.join()
            self._reraise_last_error()

    def _reraise_last_error(self):
        """Reraise any error caught in the scheduler thread.

        Resets the stored exception, so this method will not raise an error
        on subsequent calls.
        """
        if self._exception is not None:
            self._exception, e = None, self._exception  # reset exception
            raise e

    def _call_action(self) -> bool:
        try:
            self._action()
        except Exception as e:
            self._store_error_and_abort(e)

    def _should_continue(self) -> bool:
        """Check whether the autoaction should schedule the next action.

        Returns:
            bool: True if running and the stop condition is not met
        """
        try:
            return self._running and not self.stop_condition()
        except Exception as e:
            self._store_error_and_abort(e)

    def _store_error_and_abort(self, exc: Exception):
        """Store an exception and immediately stop the scheduler thread.
        This method should only be called from the scheduler thread.
        """
        self._exception = exc
        self._running = False
        sys.exit()  # immediately exit the thread

    # ================================= SCHEDULING =================================
    def _scheduler_callback(self, scheduled_time: float):
        """Calls the user action callback and schedules the next action, if any.

        This method is called periodically by the scheduler and should not be
        called directly.

        Args:
            scheduled_time (float): the time that *this* event was scheduled for.
                The next action is scheduled to occur some amount of time after
                this value.
        """
        if self._should_continue():
            self._call_action()

            next_time = self._get_next_action_time(scheduled_time)
            self._schedule_action(next_time)

    def _make_scheduler(self) -> sched.scheduler:
        """Make a scheduler and schedule the first trigger event.

        Args:
            check_continue (callable): a function that takes no arguments and returns a bool indicating
                whether or not to continue sending triggers.

        Returns:
            sched.scheduler: the new scheduler object
        """
        return sched.scheduler(timefunc=self._time_fn, delayfunc=self._delay_fn)

    def _schedule_action(self, scheduled_time: float):
        """Schedule a new trigger event to occur at the given time.

        Args:
            scheduled_time (float): the time at which the next action should occur.
                Units are in seconds, and the point of reference depends on
                the time function used in the constructor.
        """
        self._scheduler.enterabs(
            time=scheduled_time,
            priority=0,  # priority doesn't matter here
            action=self._scheduler_callback,
            argument=(scheduled_time,),
        )

    def _get_next_action_time(self, current_time: float = None) -> float:
        """Get the time that the next action should be scheduled for.

        Args:
            current_time (float): the time of the current action. If not provided,
                the value is assumed from the time function.
        """
        return (current_time or self._time_fn()) + self._interval



"""Initializer for the UPAC96 board

This initializer will run the init sequence for the UPAC96 board.
It contains all UPAC96 specific startup sequence
"""
import functools
import logging
import time

from naludaq.board import initializers
from naludaq.controllers import get_board_controller, get_dac_controller

logger = logging.getLogger(__name__)


def func_print(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger.info(f" Running: {func.__name__} ".ljust(80, "#"))
        v = func(*args, **kwargs)
        return v

    return wrapper


class InitUPAC96(initializers.Initializers):
    def __init__(self, board):
        super().__init__(board, "upac96")
        self.power_sequence = [
            "pwr_udc1_en",
            "pwr_udc2_en",
        ]
        self.rxtx = [
            "udc_rxout_enable",
            "udc_txin_enable",
        ]
        self.num_chips = self.board.params.get("num_chips", 6)

        self.conn_type = self.board.connection_info.get("type", None)
        if self.conn_type not in ["uart", "ftdi", "ft60x"]:
            raise ConnectionError(f"Unsupported connection type: {self.conn_type}")

    def run(self):
        logger.info("Starting UPAC96 with %s connection", self.conn_type)
        # Enable UART

        self._select_output_buffer()

        # UDC_FPGA_register_init
        self._fpga_init()

        # POWER ON FMC BOARD
        self._power_toggle(True)

        # reset sysrst and regclr
        self._reset_sys()

        self._reset_digital()

        # Enable Rx and Tx
        self._rxtx_toggle(False)
        time.sleep(0.1)
        self._rxtx_toggle(True)

        # Analog and digital startup:
        # Digital startup
        self._digital_startup()

        # Analog_startup
        self._analog_startup()
        self._write_chip_ids()
        # UDC ASIC reset
        # self._asic_reset()

        # Set DACS
        self._init_dacs()

        # for some reason on USB there's junk left over in the buffer
        self.board.connection.reset_input_buffer()

        return True

    @func_print
    def _fpga_init(self):
        """Write all FPGA registers to their default values."""
        self.control_registers.write_all()
        self.i2c_registers.write("i2c_en", True)

    @func_print
    def _power_toggle(self, state):
        """Toggles the power rails defined in `power_sequence` attribute"""
        for register in self.power_sequence:
            self.control_write(register, state)

    def _select_output_buffer(self, enable=False):
        """Enable the uart interface and disable the usb interface"""
        # Flush fifos, disable USB & enable UART
        if self.conn_type in ["uart", "ftdi"]:
            get_board_controller(self.board).enable_uart()
        elif self.conn_type == "ft60x":
            get_board_controller(self.board).enable_usb()
        else:
            raise ConnectionError(
                f"Connection type {self.conn_type} not supported, can't start the board"
            )

    @func_print
    def _reset_sys(self):
        """Reset the ASIC and clear the analog and digital registers"""
        self.control_write("sysrst", True)
        self.control_write("sysrst", False)
        self.control_write("regclr", True)
        self.control_write("regclr", False)

    def _reset_digital(self):
        """Reset digital side of the ASIC"""
        self.control_write("digrst", True)
        self.control_write("digrst", False)

    @func_print
    def _asic_reset(self):
        """Reset the asic, analog side, digital side, and fifo"""
        self.control_write("sysrst", True)
        self.control_write("sysrst", False)
        # self._reset_digital()

    @func_print
    def _rxtx_toggle(self, state):
        """Toggles ASIC rx and tx enable. True when talking to ASIC."""
        value = 0x3F if state else 0x00
        for register in self.rxtx:
            self.control_write(register, value)

    @func_print
    def _digital_startup(self):
        """Startup the digital side by programming all registers"""
        self.digital_registers.write_all()

    @func_print
    def _write_chip_ids(self):
        """Write the chip ids to the digital registers

        This will allow the parser to recognize which chip the return data
        belongs to.
        """
        udc_rxout_enable = (
            self.board.registers["control_registers"]
            .get("udc_rxout_enable", {})
            .get("value", 0x3F)
        )
        for chip in range(self.num_chips):
            self.control_registers.write("udc_rxout_enable", 2**chip)
            self.digital_registers.write("chip_id", chip)
        self.control_registers.write("udc_rxout_enable", udc_rxout_enable)

    @func_print
    def _analog_startup(self):
        """Start the analog side of the chip"""
        self.analog_registers.write_all()

        self._write_reg1()

        # DLL START UP SEQUENCE
        self._dll_startup()

    @func_print
    def _write_reg1(self):
        """reg1 is weird, two things live at the same address.
        Need to switch between them.
        """
        self.digital_write("writemask_r", 0x0FF)
        time.sleep(0.01)
        self.analog_write("reg1", 0x100)
        time.sleep(0.01)
        self.digital_write("writemask_r", 0x1FF)
        time.sleep(0.01)
        self.digital_write("writemask_l", 0x0FF)
        time.sleep(0.01)
        self.analog_write("reg1", 0x100)
        time.sleep(0.01)
        self.digital_write("writemask_l", 0x1FF)

    @func_print
    def _dll_startup(self):
        """Starting the delay line on the analog side."""
        self.analog_write("qbias", 0)
        self.analog_write("vadjn_sw", 1)
        time.sleep(1)
        self.analog_write("qbias", 2048)
        self.analog_write("vadjn_sw", 0)

    @func_print
    def _init_dacs(self):
        """Program the left and right side DACs"""
        dacvalues = self._get_dacvalues()

        for channel, value in dacvalues.items():
            get_dac_controller(self.board).set_single_dac(channel, value)

    def _get_dacvalues(self) -> dict:
        """Return a dict with all the "channels: dac_values"""
        return self.board.params.get("ext_dac", {}).get("channels", {})



"""Base class for registers.
"""
from abc import ABC
from collections import defaultdict
from functools import wraps
from logging import getLogger
from typing import Iterable

from naludaq.backend.exceptions import ConnectionError
from naludaq.backend.managers.connection import ConnectionManager
from naludaq.backend.managers.io import BoardIoManager
from naludaq.helpers.exceptions import InvalidRegisterError

LOGGER = getLogger(__name__)


def clear_buffer(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """Clears the board buffer before"""
        if self.board.using_new_backend:
            device = ConnectionManager(self.board).device
            if device is None:
                raise ConnectionError("There is no active connection")
            device.clear_buffers()
        else:
            self.board.connection.reset_input_buffer()
        rv = func(self, *args, **kwargs)

        return rv

    return wrapper


class Registers(ABC):
    """Base class for registers.

    By inehriting this class it will be compatible with the other register classes.
    This standardizes the way register updates are called.

    Atrributes:
    -----------

        board: The board to update the registers on.
        width_cmd (int): width of the command type in hex chars
        width_addr (int): width of the address in hex chars
        width_val (int): width of the value in hex chars
        width_total (int): total width of the command in hex chars

    """

    def __init__(self, board, register_type: str):
        self.board = board
        self._register_type = register_type
        if not isinstance(getattr(board, "registers", None), dict):
            raise InvalidRegisterError("The board is lacking valid registers")
        if not board.registers.get(register_type, None):
            raise InvalidRegisterError(f"The registers is lacking {register_type}")
        self.width_cmd = 3
        self.width_addr = 2  # hex char
        self.width_val = 3  # hex char
        self.width_total = 8  # common firmware

        self._stopword = None
        self.stopword = self.board.params.get(
            "register_stop_word",
            self.board.params["stop_word"],
        )

    @property
    def stopword(self):
        """Get/Set stopword"""
        return self._stopword

    @stopword.setter
    def stopword(self, value):
        if not isinstance(value, (int, str, bytes)):
            raise TypeError(f"stopword must be str, bytes or hex, got {type(value)}")
        elif isinstance(value, str):
            if len(value) % 2 != 0:
                raise ValueError(
                    "Since stopword converts to bytes, it must be of even length."
                )
            value = bytes.fromhex(value)
        elif isinstance(value, int):
            value = bytes.fromhex(f"{value:X}")
        self._stopword = value

    @property
    def registers(self) -> dict:
        """Shortcut to board registers"""
        return self.board.registers[self._register_type]

    @property
    def width_val_bits(self):
        """Maximum bits a value can be based on command size"""
        return self.width_val * 4

    # VALIDATIONS ############################################################
    #
    ##########################################################################
    def _validate_addr_or_raise(self, addr: str):
        """Validate the address type and make sure it exists.

        Args:
            addr(str): Hex value of address.

        Raises:
            ValueError if addr is not a valid hex or if it doesn't exist in register
            TypeError if addr is not a string or int
        """
        if not isinstance(addr, str):
            raise TypeError("Register number needs to be an integer")

        if addr.zfill(self.width_addr).upper() not in self.list_addresses():
            raise ValueError(f"Address {addr} does not exist in {self._register_type}")

    def _validate_name_or_raise(self, name):
        """Name is a user input, make sure it's valid.

        Args:
            name(str): Name of the analog register.

        Returns:
            True if name is valid, False if it's not.

        Raises:
            TypeError and ValueError
        """
        if not isinstance(name, str):
            raise TypeError(f"Name: {name} is of the wrong type: {type(name)}")
        if name.casefold() not in [k.casefold() for k in self.registers.keys()]:
            raise ValueError(f"Name: {name} is not a valid analog register.")

    def _validate_names_or_raise(self, names: Iterable[str]):
        """Validate list of register names or raise an error."""
        if not isinstance(names, Iterable):
            raise TypeError("Names must be Iterable[str]")
        for name in names:
            self._validate_name_or_raise(name)

    def _validate_value_or_raise(self, value: int, bitwidth: int):
        """Value is a user input, make sure it's valid.

        Args:
            value: Value to set the analog register to.

        Raises:
            TypeError: if the value is of the wrong type.
            ValueError: if the Value is not positive.
        """
        if not isinstance(value, (type(None), int)):
            raise TypeError(f"Value argument is of the wrong type: {type(value)}")
        if isinstance(value, int) and not 0 <= value < 2**bitwidth:
            raise ValueError("Value argument must be  a positive integer.")

    def _validate_value_map_or_raise(self, values: dict[str, int]):
        """Validate the given value map or raise an error"""
        if not isinstance(values, dict):
            raise TypeError("Values must be dict[str, int]")
        for name, value in values.items():
            self._validate_name_or_raise(name)
            bitwidth = self.registers[name]["bitwidth"]
            self._validate_value_or_raise(value, bitwidth)

    # HELPERS ################################################################
    #
    # A set of helper functions used by all registers
    #
    ##########################################################################

    def _get_addr_from_name_or_raise(self, name):
        """Returns the control registers addr from the name.

        Args:
            name(str): Name of the control register

        Returns:
            The number of the control register
        """
        if not isinstance(name, str):
            raise TypeError("Register name must be a strign")

        addr = self.registers.get(name.casefold(), False)
        if addr is False:
            raise InvalidRegisterError(f"{name} is not a valid register.")
        addr = addr["address"]
        if isinstance(addr, int):
            addr = f"{addr:X}"  # pragma: no cover
        addr = addr.zfill(self.width_addr)
        return addr.upper()

    def _validate_command_or_raise(self, command: str):
        """Make sure a generated command is of the right length."""
        if len(command) != self.width_total:
            raise ValueError(
                f"Generated command is invalid. A command must be {self.width_total} char, got {command} or length {len(command)}"
            )  # pragma: no cover

    # COMMON #################################################################
    #
    # A list of convenience tools, common for all registers.
    #
    ##########################################################################
    def list(self) -> list:
        """Lists the name of all available registers.

        Returns:
            A list of all register names
        """
        return list(k.lower() for k in self.registers.keys())

    def list_registers(self) -> dict:
        """Returns all available registers as a dict, organized in the
        same way as the registers file.

        Returns:
            A dict containing all registers
        """
        return self.registers

    def list_addresses(self):
        """Generate a list of all available addresses.

        This is useful since an address can contain multiple registers.
        The register name can correspond to one or many bits at a specific address.

        Returns:
            List of available addresses in hex (str)
        """
        reg_addr = {value["address"].upper() for value in self.registers.values()}
        return sorted(list(reg_addr), key=lambda x: int(x, 16))

    def show(self, name: str) -> str:
        """Returns the internal contents of a register.

        Args:
            name (str): The name of the register

        Returns:
            A string representation of the given register.

        Raises:
            TypeError if the name is not a string
            ValueError if the name is not a valid register
        """
        self._validate_name_or_raise(name)

        return self.registers[name]

    def _send_command(self, cmd):
        """Send request to the board"""
        if self.board.using_new_backend:
            BoardIoManager(self.board).write(cmd)
        else:
            self.board.connection.send(cmd)

    def __repr__(self):
        return f"{self.registers}(board = {self.board})"  # pragma: no cover


class WriteRegisters(Registers):
    """Writeable registers

    Adds the trait to write registers

    Functions:
    ---------

        - ``write(name, value)``
        - ``write_addr(addr, value)``
        - ``generate_write(name, value)``
        - ``generate_write_addr(name, value)``
        - ``write_all()``

    Attributes:
    -----------

        _write_cmd: the write command as a hex eg. ``0xAD``

    """

    _write_cmd = None

    # WRITE ##################################################################
    def write(self, name: str, value: int = None):
        """Set a register to a value.

        Update the internal stored reg values and
        the registers on the hardware simultaniously.

        Args:
            board
            name(str): Name of the register.
            value(int, bool): Value of the register.

        Raises:
            ValueError if name is not in register.
            TypeError if name is not a str.
        """
        self._validate_name_or_raise(name)
        name = name.lower()
        if value is not None:
            self._update_software_register(name, value)

        addr = self._get_addr_from_name_or_raise(name)

        self._write_register_or_raise(addr)

    def write_many(self, value_map: dict[str, int]):
        """Write several registers at once.

        Args:
            registers (dict[str, int]): mapping of register values to names
        """
        self._validate_value_map_or_raise(value_map)

        # update all software registers first to prevent generate_write
        # from using potentially old register values in subsequent writes
        for name, value in value_map.items():
            self._update_software_register(name, value)

        addresses = {
            self._get_addr_from_name_or_raise(name) for name in value_map.keys()
        }
        commands = [self.generate_write_addr(addr) for addr in addresses]
        command = "".join(commands)
        self._send_command(command)

    def write_addr(self, addr: "int|str", value=None):
        """Updates value on both board from register addr.

        You better know what you are doing if you use this function since it'll
        overwrite the entire address.

        This means several registers might get overwritten!

        If a value is left out the function will use whatever value is previously stored.
        The intent is to be able to write to registers that have no name yet.

        Args:
            addr (hex or int): the addr can either be hex or integer.
            value: the value sent to the board.

        Raises:
            ValueError if value is not a positive integer
            TypeError if name is not a str.
        """
        if isinstance(addr, int):
            addr = f"{addr:X}"
        self._validate_addr_or_raise(addr)
        addr = addr.zfill(self.width_addr)
        addr = addr.upper()  # All hex is uppercase

        if value is not None:
            self._update_software_register_addr(addr, value)

        self._write_register_or_raise(addr)

    def write_all(self):
        """Write all software registers to the hardware.

        Writes all registers on the board-object to the physical board.
        It will send all the commands one by one to the board.
        Effectively syncing the software registers with the board by
        sending all values in the software to the board.
        """
        if self.board.using_new_backend:
            return self._write_all_new()
        for addr in self.list_addresses():
            self._write_register_or_raise(addr)

    def _write_all_new(self, batch_size: int = 32):
        """Version of the write_all command that writes commands simulataneously.

        Args:
            batches (int): number of commands per batch to send.
                Don't make this too high, since serial/d2xx connections are
                synchronous and so large batches can result in a request timeout
                before the server finishes writing.
        """
        addresses = self.list_addresses()
        command_batches = [
            [self.generate_write_addr(addr) for addr in addresses[i : i + batch_size]]
            for i in range(0, len(addresses), batch_size)
        ]
        for commands in command_batches:
            BoardIoManager(self.board).write_all(commands)

    def _write_register_or_raise(self, addr):
        """Write a control register to the board using the register addr.

        Uses the value stored in the software and sends to the hardware.

        Args:
            addr(str): The hex address for the register as str.

        Raises:
            TypeError: if addr is not an str
            ValueError: if reg_num is not a valid register.
        """
        addr = addr.upper()
        self._validate_addr_or_raise(addr)
        value = self._get_value_from_addr(addr)

        cmd = self.generate_write_addr(addr, value)

        self._send_command(cmd)

    # GENERATE ###############################################################
    #
    # A series of commands the generates the commands to send and returns a hex str.
    #
    ##########################################################################
    def generate_write(self, name: str, value: int = None) -> str:
        """Generates a command string for writing to a register.

        This function does not change any hardware or software registers.

        Args:
            name (str): the name of the register
            value (int): the value, or None to use the value held in the software register.

        Returns:
            The command string

        Raises:
            ValueError if the address, value or generated command is invalid
            TypeError if the value is an invalid type
        """
        self._validate_name_or_raise(name)
        name = name.lower()
        addr = self._get_addr_from_name_or_raise(name)
        # Fix total comand for address

        if value is not None:
            addr_val = int(self._get_value_from_addr(addr))
            bw = self.registers[name]["bitwidth"]
            bp = self.registers[name]["bitposition"]
            self._validate_value_or_raise(value, bw)
            mask = (2**bw - 1) << bp
            value = addr_val - (addr_val & mask) + ((value & (2**bw - 1)) << bp)

        try:
            cmd = self.generate_write_addr(addr=addr, value=value)
        except (ValueError, TypeError, AttributeError):  # pragma: no cover
            raise  # pragma: no cover
        return cmd

    def generate_write_addr(self, addr: "int|str", value: int = None) -> str:
        """Generates a command string for writing to a register.

        This function does not change any hardware or software registers.

        Args:
            addr (str): the address of the register
            value (int): the value, or None to use the value held in the software register.

        Returns:
            The command string

        Raises:
            ValueError if the address, value or generated command is invalid
            TypeError if the value is an invalid type
        """
        if isinstance(addr, int):
            addr = f"{addr:X}"
        if not isinstance(addr, str):
            raise TypeError("address should be either a hex string or hex (0xADDR)")
        addr = addr.zfill(self.width_addr)
        addr = addr.upper()

        if value is None:
            try:
                value = self._get_value_from_addr(addr)
            except AttributeError:
                raise ValueError(f"No value found for address {addr}")

        self._validate_value_or_raise(value, bitwidth=self.width_val_bits)

        command = self._generate_write_request(addr, value)
        self._validate_command_or_raise(command)

        return command

    def _generate_write_request(self, addr: str, value: int) -> str:
        """Generate the write command"""

        cmd = f"{self._write_cmd:X}"
        cmd_addr = addr.zfill(self.width_addr)
        val = f"{value:X}".zfill(self.width_val)
        command = f"{cmd}{cmd_addr}{val}"
        return command.upper()

    def _update_software_register(self, name: str, value: int):
        """Set boardparam digital register, do not change the state of the FPGA

        Args:
            name: Register name.
            value: Register value.
        """
        name = name.lower()
        self._validate_name_or_raise(name)
        self._validate_value_or_raise(value, bitwidth=self.registers[name]["bitwidth"])

        self.registers[name]["value"] = value

    def _update_software_register_addr(self, addr: str, value: int):
        """Update any registers sharing the addr with their part of the value"""
        self._validate_value_or_raise(
            value, bitwidth=self.width_val_bits
        )  # hex is 4 bits
        for reg in self.list_registers().values():
            if reg["address"].lower() != addr.lower():
                continue
            val = value >> reg["bitposition"]
            val &= 2 ** reg["bitwidth"] - 1
            reg["value"] = val

    def _get_value_from_addr(self, addr):
        """Generate the value from the hex address.

        An address can contain multiple registers.
        Generate the value depending on the bit positions of all registers on the address.

        Args:
            addr(str): Hex value of the address, 2 hex char long.

        Returns:
            integer value of the combined registers.

        Raises:
            AttributeError if address is not in the registers.
        """
        regs = [
            val
            for val in self.registers.values()
            if val["address"].upper() == addr.upper()
        ]
        if len(regs) == 0:
            raise AttributeError("Address not found")

        value = 0
        for val in regs:
            # key_value, key_addr, key_pos, key_width, _ = val
            key_value = val["value"]
            val["address"]
            key_pos = val["bitposition"]
            key_width = val["bitwidth"]

            # if key_addr == addr:
            value += 2**key_pos * (int(key_value) % (2**key_width))

        return value


class ReadRegisters(Registers):
    """Readable registers

    Adds the ability to read registers


    Functions:
    ---------

        - ``read(name)``
        - ``read_addr(addr)``
        - ``generate_read(name)``
        - ``generate_read_addr(name)``
        - ``read_all()``

    Attributes:
    -----------

        _read_cmd: the read address as a hex eg. ``0xAF``

    """

    _read_cmd = None

    @clear_buffer
    def read(self, name: str) -> int:
        """Read from a register with a name.

        Args:
            name: Name of the register.

        Raises:
            ValueError if name is not in register.
            TypeError if name is not a str.
        """
        self._validate_name_or_raise(name)
        name = name.lower()
        reg_addr = self._get_addr_from_name_or_raise(name)

        return_value = self._read_from_hardware(reg_addr)
        reg = self._parse_response_to_names(return_value, reg_addr)

        return reg[name]

    @clear_buffer
    def read_many(self, names: list[str]) -> list[str]:
        """Read several registers at once.

        Do not use this function to read the same register multiple times,
        as older values will be discarded.

        Args:
            names (Iterable[str]): names of registers to read

        Returns:
            list[int]: list of register values in the same order as provided.

        Raises:
            TimeoutError: if one or more registers could not be read in time.
        """
        self._validate_names_or_raise(names)

        # cannot perform concurrent reads without naludaq_rs
        if not self.board.using_new_backend:
            return [self.read(name)["value"] for name in names]

        # sort registers & generate by address to reduce # of commands
        addresses = defaultdict(list)
        for name in names:
            addr = self._get_addr_from_name_or_raise(name)
            addresses[addr].append(name)
        read_cmds = [self.generate_read_addr(addr) for addr in addresses.keys()]

        # response order is guaranteed to be the same as the commands sent
        try:
            responses = BoardIoManager(self.board).read_all(read_cmds)
        except TimeoutError:
            raise

        # parse responses into registers by address
        values = {}
        for (addr, addr_names), response in zip(addresses.items(), responses):
            response = self._parse_answer(response)["value"]
            response = self._parse_response_to_names(response, addr)
            for name in addr_names:
                values[name] = response[name]["value"]

        sorted_values = [values[name] for name in names]
        return sorted_values

    @clear_buffer
    def read_all(self, overwrite: bool = False):
        """Read all digital registers from the board.

        Reads the values based on the register number rather than the name.

        Args:
            board

        Returns:
            dictionary with {reg_num: values,}
        """
        if self.board.using_new_backend:
            return self._read_all_new()
        values = dict()
        for reg_addr in self.list_addresses():
            response = self._read_from_hardware(reg_addr)

            reg = self._parse_response_to_names(response, reg_addr)
            values.update(reg)
        return values

    def _read_all_new(self):
        """Read all registers [BETA].

        Sends read commands all at once and parses the results.

        Returns:
            dict: register values structured as {reg_name: reg dict}
        """
        addresses = self.list_addresses()
        commands = [self.generate_read_addr(addr) for addr in addresses]
        values = {}
        answers = BoardIoManager(self.board).read_all(commands)
        for address, answer in zip(addresses, answers):
            parsed_answer = self._parse_answer(answer)["value"]
            reg = self._parse_response_to_names(parsed_answer, address)
            values.update(reg)
        return values

    @clear_buffer
    def read_addr(self, addr: int) -> int:
        """Read an address and return the value.

        Args:
            addr(hex|int): 8-bit address to read as either int or hex."""
        addr = f"{addr:02X}".upper()
        return self._read_from_hardware(addr)

    def generate_read(self, name: str) -> str:
        """Generate and return the read command as hex"""
        self._validate_name_or_raise(name)
        name = name.lower()
        addr = self._get_addr_from_name_or_raise(name)
        cmd = self._generate_read_request(addr)
        return cmd

    def generate_read_addr(self, addr: "int|str") -> str:
        """Generate and return the read addr command as HEX"""
        if isinstance(addr, int):
            addr = f"{addr:X}"
        self._validate_addr_or_raise(addr)
        addr = addr.upper()
        cmd = self._generate_read_request(addr)
        return cmd.upper()

    @clear_buffer
    def _read_from_hardware(self, addr: str, *args, **kwargs) -> int:
        """Read from a digital read register with an address

        Args:
            addr: Address to read out data from.
        """
        cmd = self._generate_read_request(addr)
        try:
            if self.board.using_new_backend:
                response = BoardIoManager(self.board).read(cmd)
            else:
                self._send_command(cmd)
                response = self._read_response()
        except TimeoutError:  # pragma: no cover
            return -1  # pragma: no cover
        answer = self._parse_answer(response)
        return answer["value"]

    def _generate_read_request(self, addr: str) -> str:
        """Generate the read command for the firmwares. cmd: RRAA000"""
        cmd = f"{self._read_cmd:X}"
        cmd_addr = addr.zfill(self.width_addr)
        val = "0".zfill(self.width_val)
        command = f"{cmd}{cmd_addr}{val}"
        return command.upper()

    def _parse_response_to_names(self, buffer, addr):
        """Takes the input buffer and parses the response into the registers.

        Each register address can contain multiple register names, this function
        parses the register address into the correct names.

        Returns:
            Dictionary with the names: [val, addr, pos, width, rw]
        """
        outp = {}

        for key, val in self.registers.items():
            if val["address"].upper() == addr.upper():
                position = val["bitposition"]
                width = val["bitwidth"]
                mask = 2**width - 1
                shifted = buffer >> position
                oval = shifted & mask
                outp[key] = {}

                for regkey, regval in val.items():
                    outp[key][regkey] = regval
                outp[key]["value"] = oval

        return outp

    def _parse_answer(self, buffer):
        """Parses a raw binary answer into python data formats.

        Args:
            buffer (bytes): Response to parse.

        Returns:
            parsed answer as a dict expected format.
        """
        header = int.from_bytes(buffer[0:2], byteorder="big", signed=False)
        response = {
            "header": header,
            "read_reg": header & 255,  # only 8 lsb]
            "value": int.from_bytes(buffer[2:4], byteorder="big", signed=False),
        }

        return response

    def _read_response(self):
        """Read a data package based on the register stopword"""
        return self.board.connection.read_until(self.stopword)  # pragma: no cover


# validation #################################################################
#
##############################################################################
def validate_registermap_or_raise(registers):
    """Validate a regmap, raises error if it's not valid."""

    # invert registers
    inv_register = invert_regmap(registers)

    # find potential conflicts
    try:
        conflicts = find_conflicting_registers(inv_register)
    except Exception as e_msg:
        raise InvalidRegisterError(f"Following register have conflicting bits: {e_msg}")
    # if potential conflicts
    if conflicts:
        raise InvalidRegisterError(
            f"Following register have conflicting bits: {conflicts}"
        )

    return True


def find_conflicting_registers(registers):
    """Takes an inverted regmap and finds all conflicts

    Args:
        registers (dict): register dictionary and find all conflicts.

    Returns:
        dictionary with conflictingaddresses and conflicting names.
    """
    conflict_regs = defaultdict(list)
    inv_regmap = invert_regmap(registers)  # {addr: [(name, reg)]}
    for addr, regs in inv_regmap.items():
        if len(regs) <= 1:
            continue

        regs = sorted(regs, key=lambda i: i[1]["bitposition"])
        for idx, (name1, item1) in enumerate(regs[:-1]):
            item1_end = item1["bitposition"] + item1["bitwidth"] - 1

            for name2, item2 in regs[idx + 1 :]:
                if item2["bitposition"] > item1_end:
                    break
                conflict_regs[addr].append(name1)
                conflict_regs[addr].append(name2)

    conflict_regs = {addr: sorted(set(names)) for addr, names in conflict_regs.items()}
    return conflict_regs


def invert_regmap(registers):
    """Invert a regmap dict to return addr: register instead of name: register

    Args:
        registers (dict): register map {name: registers}

    Returns:
        Dictionary with {address: register}
    """
    all_regs = defaultdict(list)

    for key, val in registers.items():
        addr = val["address"]
        all_regs[addr].append((key, val))

    return all_regs
